"""Data Processing and Plotting for Scenario Comparison."""

import os
from typing import Dict, Any, Tuple, Optional, List
import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from ordered_set import OrderedSet
import seaborn as sns

from hisim.postprocessing.scenario_evaluation.result_data_processing import (
    ResultDataTypeEnum,
    ResultDataProcessingModeEnum,
    ScenarioDataProcessing,
    DataFormatEnum,
)
from hisim.postprocessing.chartbase import ChartFontsAndSize
from hisim import log


class ScenarioChartGeneration:

    """ScenarioChartGeneration class."""

    def __init__(
        self,
        simulation_duration_to_check: str,
        filepath_of_aggregated_dataframe: str,
        data_processing_mode: str,
        data_format_type: str,
        scenario_config_name: str,
        time_resolution_of_data_set: str,
        dict_with_extra_information_for_specific_plot: Dict[str, Dict],
        variables_to_check: Optional[List[str]] = None,
        dict_of_scenarios_to_check: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        """Initialize the class."""

        warnings.filterwarnings("ignore")

        self.show_plot_legend: bool = True
        self.data_processing_mode = data_processing_mode
        self.path_addition: str = ""
        self.plot_path_complete: str = ""
        self.data_format_type: str = data_format_type
        self.scenario_config_name: str = scenario_config_name

        if self.data_processing_mode == ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name:
            result_path_strip = "results_all_parameters"
            self.show_plot_legend = False

        elif self.data_processing_mode == ResultDataProcessingModeEnum.PROCESS_FOR_DIFFERENT_BUILDING_CODES.name:
            result_path_strip = "results_different_building_codes"

        else:
            raise ValueError("DataProcessingMode not known.")

        self.result_folder = os.path.join(
            os.getcwd(),
            os.pardir,
            os.pardir,
            os.pardir,
            "system_setups",
            "scenario_comparison",
            "results",
            result_path_strip,
            f"{simulation_duration_to_check}_days",
        )
        if not os.path.exists(self.result_folder):
            os.makedirs(self.result_folder)

        self.hisim_chartbase = ChartFontsAndSize()
        self.hisim_chartbase.figsize = (10, 6)
        self.hisim_chartbase.dpi = 100

        if variables_to_check != [] and variables_to_check is not None:
            # read data, sort data according to scenarios if wanted, and create pandas dataframe

            (
                full_dataframe,
                variables_to_check,
            ) = ScenarioDataProcessing.get_dataframe_and_create_pandas_dataframe_for_all_data(
                filepath_of_aggregated_dataframe=filepath_of_aggregated_dataframe,
                dict_of_scenarios_to_check=dict_of_scenarios_to_check,
                variables_to_check=variables_to_check,
                data_format_type=self.data_format_type,
            )

            self.make_plots_with_specific_kind_of_data(
                time_resolution_of_data_set=time_resolution_of_data_set,
                full_dataframe=full_dataframe,
                simulation_duration_key=simulation_duration_to_check,
                variables_to_check=variables_to_check,
                dict_with_extra_information_for_specific_plot=dict_with_extra_information_for_specific_plot,
            )

        else:
            log.information("Variable list for data is not given and will not be plotted or anaylzed.")

    def make_plots_with_specific_kind_of_data(
        self,
        time_resolution_of_data_set: str,
        full_dataframe: pd.DataFrame,
        simulation_duration_key: str,
        variables_to_check: List[str],
        dict_with_extra_information_for_specific_plot: Dict[str, Dict],
    ) -> None:
        """Make plots for different kind of data."""

        log.information(f"Simulation duration: {simulation_duration_key} days.")

        if full_dataframe.empty:
            raise ValueError("Dataframe is empty.")

        sub_results_folder = f"{time_resolution_of_data_set}_{self.scenario_config_name}"

        self.path_for_plots = os.path.join(self.result_folder, sub_results_folder)

        for variable_to_check in variables_to_check:
            log.information(f"Checking variable: {variable_to_check}")

            self.prepare_plot_path(variable_to_check)

            filtered_data = self.filter_dataframe_according_to_output_variable(full_dataframe, variable_to_check)
            x_and_y_plot_data, output_value_keys = self.get_mean_values(
                filtered_data=filtered_data, time_resolution_of_data_set=time_resolution_of_data_set
            )

            if self.check_empty_output_values(filtered_data, output_value_keys):
                continue

            self.save_filtered_data_to_excel_or_csv(filtered_data, data_format_type=self.data_format_type)

            unit = self.get_unit(filtered_data=filtered_data, variable_to_check=variable_to_check)

            if time_resolution_of_data_set == ResultDataTypeEnum.YEARLY.name:
                self.process_yearly_data(
                    filtered_data=filtered_data,
                    full_dataframe=full_dataframe,
                    x_and_y_plot_data=x_and_y_plot_data,
                    output_value_keys=output_value_keys,
                    unit=unit,
                    variable_to_check=variable_to_check,
                    dict_with_extra_information_for_specific_plot=dict_with_extra_information_for_specific_plot,
                )
            elif time_resolution_of_data_set in (
                ResultDataTypeEnum.HOURLY.name,
                ResultDataTypeEnum.DAILY.name,
                ResultDataTypeEnum.MONTHLY.name,
            ):
                self.process_time_series_data(
                    filtered_data=filtered_data,
                    full_dataframe=full_dataframe,
                    x_and_y_plot_data=x_and_y_plot_data,
                    output_value_keys=output_value_keys,
                    unit=unit,
                    variable_to_check=variable_to_check,
                    dict_with_extra_information_for_specific_plot=dict_with_extra_information_for_specific_plot,
                    time_resolution_of_data_set=time_resolution_of_data_set,
                )
            else:
                raise ValueError("Invalid time resolution.")

    def prepare_plot_path(self, variable_to_check: str) -> None:
        """Prepare path for saving plots."""
        if "|" in variable_to_check:
            self.path_addition = f"{variable_to_check.split('|')[0]}_{variable_to_check.split('|')[-1]}"
        else:
            self.path_addition = variable_to_check.replace(" ", "_")

        self.plot_path_complete = os.path.join(self.path_for_plots, self.path_addition)
        os.makedirs(self.plot_path_complete, exist_ok=True)

    def filter_dataframe_according_to_output_variable(
        self, dataframe: pd.DataFrame, variable_to_check: str
    ) -> pd.DataFrame:
        """Filter dataframe by variable."""
        filtered_dataframe = ScenarioDataProcessing.filter_pandas_dataframe_according_to_output_variable(
            dataframe=dataframe, variable_to_check=variable_to_check
        )
        return filtered_dataframe

    def get_mean_values(
        self, filtered_data: pd.DataFrame, time_resolution_of_data_set: str
    ) -> Tuple[pd.DataFrame, List[str]]:
        """Calculate mean values of scenarios."""
        x_and_y_plot_data, keys_for_output_values = ScenarioDataProcessing.take_mean_values_of_scenarios(
            filtered_data=filtered_data, time_resolution_of_data_set=time_resolution_of_data_set
        )
        return x_and_y_plot_data, keys_for_output_values

    def check_empty_output_values(self, filtered_data: pd.DataFrame, output_value_keys: List[str]) -> bool:
        """Check if any output column has no values."""
        for output_value_key in output_value_keys:
            if filtered_data[("Output", output_value_key)].isnull().all():
                log.information(f"No values for {output_value_key} found.")
                return True
        return False

    def save_filtered_data_to_excel_or_csv(self, filtered_data: pd.DataFrame, data_format_type: str) -> None:
        """Save filtered data to Excel or CSV."""
        if data_format_type == DataFormatEnum.XLSX.name:
            filtered_data.to_excel(os.path.join(self.plot_path_complete, "results.xlsx"))
        elif data_format_type == DataFormatEnum.CSV.name:
            filtered_data.to_csv(os.path.join(self.plot_path_complete, "results.csv"))
        else:
            raise ValueError(f"Only data format types xlsx or csv are implemented. Here it is {self.data_format_type}.")

    def get_unit(self, filtered_data: pd.DataFrame, variable_to_check: str) -> str:
        """Get unit of variable."""
        try:
            return str(filtered_data[("Output", "unit")].values[0])
        except Exception:
            return "°C*h" if "Temperature deviation" in variable_to_check else "-"

    def process_yearly_data(
        self,
        full_dataframe: pd.DataFrame,
        filtered_data: pd.DataFrame,
        x_and_y_plot_data: pd.DataFrame,
        output_value_keys: List[str],
        unit: str,
        variable_to_check: str,
        dict_with_extra_information_for_specific_plot: Dict[str, Dict],
    ) -> None:
        """Process yearly data."""
        kind_of_data_set = "yearly"
        self.get_statistics_and_write_to_excel(filtered_data, kind_of_data_set, x_and_y_plot_data)

        self.make_box_plot(
            filtered_data=filtered_data,
            unit=unit,
            key_for_output_values=output_value_keys[0],
            y_axis_label=self.path_addition.replace("_", " "),
        )

        self.make_bar_plot(
            unit=unit, x_and_y_plot_data=x_and_y_plot_data, y_axis_label=self.path_addition.replace("_", " ")
        )

        try:
            x_data_variable = dict_with_extra_information_for_specific_plot["scatter"]["x_data_variable"]

            self.make_scatter_plot_for_yearly_data(
                full_pandas_dataframe=full_dataframe,
                filtered_data=filtered_data,
                y_data_variable=self.path_addition.replace("_", " "),
                x_data_variable=x_data_variable,
                key_for_output_values=output_value_keys[0],
            )
        except Exception:
            log.information(f"{variable_to_check} could not be plotted as scatter plot.")

        try:
            self.make_histogram_plot(
                filtered_data=filtered_data,
                key_for_output_values=output_value_keys[0],
                unit=unit,
                x_axis_label=self.path_addition.replace("_", " "),
            )
        except Exception:
            log.information(f"{variable_to_check} could not be plotted as histogram plot.")

        if variable_to_check in [
            dict_with_extra_information_for_specific_plot["stacked_bar"]["y1_data_variable"],
            dict_with_extra_information_for_specific_plot["stacked_bar"]["y2_data_variable"],
        ]:
            y1_data_variable = dict_with_extra_information_for_specific_plot["stacked_bar"]["y1_data_variable"]
            y2_data_variable = dict_with_extra_information_for_specific_plot["stacked_bar"]["y2_data_variable"]
            use_y1_as_bottom_for_y2 = dict_with_extra_information_for_specific_plot["stacked_bar"][
                "use_y1_as_bottom_for_y2"
            ]
            sort_according_to_y1_or_y2_data = dict_with_extra_information_for_specific_plot["stacked_bar"][
                "sort_according_to_y1_or_y2_data"
            ]
            self.make_stacked_bar_plot(
                full_pandas_dataframe=full_dataframe,
                y1_data_variable=y1_data_variable,
                y2_data_variable=y2_data_variable,
                use_y1_as_bottom_for_y2=use_y1_as_bottom_for_y2,
                sort_according_to_y1_or_y2_data=sort_according_to_y1_or_y2_data,
            )

    def process_time_series_data(
        self,
        filtered_data: pd.DataFrame,
        full_dataframe: pd.DataFrame,
        x_and_y_plot_data: pd.DataFrame,
        output_value_keys: List[str],
        unit: str,
        variable_to_check: str,
        dict_with_extra_information_for_specific_plot: Dict[str, Dict],
        time_resolution_of_data_set: str,
    ) -> None:
        """Process time series data."""
        if time_resolution_of_data_set == ResultDataTypeEnum.HOURLY.name:
            kind_of_data_set = "hourly"
            line_plot_marker_size = 2
        elif time_resolution_of_data_set == ResultDataTypeEnum.DAILY.name:
            kind_of_data_set = "daily"
            line_plot_marker_size = 3
        elif time_resolution_of_data_set == ResultDataTypeEnum.MONTHLY.name:
            kind_of_data_set = "monthly"
            line_plot_marker_size = 5

        self.get_statistics_and_write_to_excel(filtered_data, kind_of_data_set, x_and_y_plot_data)

        self.make_line_plot(
            unit=unit,
            x_and_y_plot_data=x_and_y_plot_data,
            line_plot_marker_size=line_plot_marker_size,
            y_axis_label=self.path_addition.replace("_", " "),
        )

        try:
            x_data_variable = dict_with_extra_information_for_specific_plot["scatter"]["x_data_variable"]
            self.make_line_scatter_plot(
                full_pandas_dataframe=full_dataframe,
                filtered_data=filtered_data,
                keys_for_output_values=output_value_keys,
                y_data_variable=self.path_addition,
                x_data_variable=x_data_variable,
                line_plot_marker_size=line_plot_marker_size,
            )
        except Exception:
            log.information(f"{variable_to_check} could not be plotted as line scatter plot.")

    def get_statistics_and_write_to_excel(
        self, filtered_data: pd.DataFrame, kind_of_data_set: str, x_and_y_plot_data: pd.DataFrame
    ) -> None:
        """Get statistical data and write to Excel."""
        ScenarioDataProcessing.get_statistics_of_data_and_write_to_excel(
            filtered_data=filtered_data,
            path_to_save=self.plot_path_complete,
            kind_of_data_set=kind_of_data_set,
            x_and_y_plot_data=x_and_y_plot_data,
        )

    def make_line_plot(
        self, x_and_y_plot_data: pd.DataFrame, line_plot_marker_size: int, y_axis_label: str, unit: str
    ) -> None:
        """Make line plot."""
        log.information("Make line plot with data.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)
        x_data = x_and_y_plot_data["time"]
        x_data_transformed = np.asarray(x_data, dtype="datetime64[D]")

        color, edgecolor = self.set_plot_colors_according_to_data_processing_mode(
            number_of_scenarios=len(x_and_y_plot_data.columns[1:]), data_processing_mode=self.data_processing_mode,
        )
        del edgecolor
        # make one line for each scenario
        for index, scenario in enumerate(x_and_y_plot_data.columns[1:]):
            plt.plot(
                x_data_transformed,
                x_and_y_plot_data[scenario],
                "-o",
                markersize=line_plot_marker_size,
                label=scenario,
                color=color[index],
            )

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            y_axis_unit=unit,
            show_legend=self.show_plot_legend,
            plot_type_name="line_plot",
            rotate_x_ticks=True,
            y_axis_label=y_axis_label,
        )

    def make_bar_plot(self, x_and_y_plot_data: pd.DataFrame, unit: str, y_axis_label: str) -> None:
        """Make bar plot."""
        log.information("Make bar plot.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)
        x_data: Any
        y_data = x_and_y_plot_data.iloc[0, 1:]
        bar_labels = x_and_y_plot_data.columns[1:]

        # sort y_data and labels
        y_data_sorted, bar_labels_sorted = self.sort_y_values_according_to_data_processing_mode(
            data_processing_mode=self.data_processing_mode, zip_list_one=list(y_data), zip_list_two=list(bar_labels)
        )
        # if no scenarios chosen, make artificial x ticks
        if self.data_processing_mode == ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name:
            x_data = np.arange(0, len(y_data) * 2, step=2)
            rotate_x_ticks = False
            x_axis_label = ""
            show_x_ticks = False
        # otherwise choose scenarios as x-ticks
        else:
            x_data = bar_labels_sorted
            rotate_x_ticks = True
            x_axis_label = ""
            show_x_ticks = True

        color, edgecolor = self.set_plot_colors_according_to_data_processing_mode(
            number_of_scenarios=len(bar_labels), data_processing_mode=self.data_processing_mode
        )
        a_x.bar(x_data, y_data_sorted, label=bar_labels_sorted, color=color, edgecolor=edgecolor)

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            show_legend=False,
            plot_type_name="bar_plot",
            y_axis_unit=unit,
            x_axis_label=x_axis_label,
            show_x_ticks=show_x_ticks,
            rotate_x_ticks=rotate_x_ticks,
            x_ticks=np.arange(len(bar_labels)),
            x_tick_labels=bar_labels_sorted,
            y_axis_label=y_axis_label,
        )

    def make_box_plot(
        self,
        filtered_data: pd.DataFrame,
        key_for_output_values: str,
        unit: str,
        y_axis_label: str,
        scenario_set: Optional[List[str]] = None,
    ) -> None:
        """Make box plot."""
        log.information("Make box plot.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)
        if scenario_set is None:
            scenario_set = list(OrderedSet(filtered_data[("Input", "scenario")]))

        sns.boxplot(
            data=filtered_data, x=("Input", "scenario"), y=("Output", key_for_output_values), palette="Spectral"
        )
        a_x.set(xlabel=None)

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            x_axis_label="",
            y_axis_unit=unit,
            show_legend=False,
            legend_labels=scenario_set,
            plot_type_name="box_plot",
            show_x_ticks=True,
            x_ticks=np.arange(len(scenario_set)),
            x_tick_labels=scenario_set,
            rotate_x_ticks=True,
            y_axis_label=y_axis_label,
        )

    def make_histogram_plot(
        self,
        filtered_data: pd.DataFrame,
        key_for_output_values: str,
        unit: str,
        x_axis_label: str,
        scenario_set: Optional[List[str]] = None,
    ) -> None:
        """Make histogram plot."""
        log.information("Make histogram plot.")

        fig, a_x = plt.subplots(  # pylint: disable=unused-variable
            figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi
        )
        if scenario_set is None:
            scenario_set = list(OrderedSet(filtered_data[("Input", "scenario")]))

        plt.hist(x=np.array(filtered_data[("Output", key_for_output_values)].values), bins="auto")

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            x_axis_label=x_axis_label,
            y_axis_label="Count",
            x_axis_unit=unit,
            show_legend=False,
            plot_type_name="histogram_plot",
        )

    def make_scatter_plot_for_yearly_data(
        self,
        full_pandas_dataframe: pd.DataFrame,
        filtered_data: pd.DataFrame,
        key_for_output_values: str,
        y_data_variable: str,
        x_data_variable: str = "Specific heating demand according to TABULA",
    ) -> None:
        """Make scatter plot."""
        log.information("Make scatter plot with data.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)

        # iterate over all scenarios
        x_data_mean_value_list_for_all_scenarios = []
        y_data_mean_value_list_for_all_scenarios = []
        for scenario in list(OrderedSet(list(full_pandas_dataframe[("Input", "scenario")]))):
            full_data_per_scenario = full_pandas_dataframe.loc[full_pandas_dataframe[("Input", "scenario")] == scenario]
            filtered_data_per_scenario = filtered_data.loc[filtered_data[("Input", "scenario")] == scenario]

            # get x_data_list by filtering the df according to x_data_variable and then by taking values from "value" column
            x_data_list = list(
                full_data_per_scenario.loc[full_data_per_scenario[("Output", "variable")] == x_data_variable][
                    ("Output", key_for_output_values)
                ].values
            )
            x_data_unit = full_data_per_scenario.loc[full_data_per_scenario[("Output", "variable")] == x_data_variable][
                ("Output", "unit")
            ].values[0]

            # if x_data_list has more than 1 value (because more values for this scenario exist), then take mean value
            if len(x_data_list) > 1:
                # for each scenario take the mean value
                x_data_mean_value_per_scenario = np.mean(x_data_list)
            elif len(x_data_list) == 1:
                x_data_mean_value_per_scenario = x_data_list[0]
            else:
                raise ValueError(
                    "The x_data_list is empty. Probably the full dataframe did not contain the x_data_variable in the column variable."
                )

            # append to x_data_mean_value_list
            x_data_mean_value_list_for_all_scenarios.append(x_data_mean_value_per_scenario)

            # get y values from filtered data per scenario (already filtered according to variable to check and scenario)
            y_data_list = list(filtered_data_per_scenario[("Output", key_for_output_values)].values)
            y_data_unit = filtered_data_per_scenario[("Output", "unit")].values[0]
            # if y_data_list has more than 1 value (because more values for this scenario exist), then take mean value
            if len(y_data_list) > 1:
                # for each scenario take the mean value
                y_data_mean_value_per_scenario = np.mean(y_data_list)
            elif len(y_data_list) == 1:
                y_data_mean_value_per_scenario = y_data_list[0]
            else:
                raise ValueError(
                    "The y_data_list is empty. Something went wrong with the filtering in the functions before."
                )

            # append to y_data_mean_value_list
            y_data_mean_value_list_for_all_scenarios.append(y_data_mean_value_per_scenario)

        # identify marker size accroding to data length
        scatter_plot_marker_size = self.get_scatter_marker_size(
            data_length=len(x_data_mean_value_list_for_all_scenarios)
        )

        # make scatter plot
        plt.scatter(
            x_data_mean_value_list_for_all_scenarios,
            y_data_mean_value_list_for_all_scenarios,
            s=scatter_plot_marker_size,
        )

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            x_axis_label=x_data_variable,
            y_axis_label=y_data_variable,
            x_axis_unit=x_data_unit,
            y_axis_unit=y_data_unit,
            show_legend=self.show_plot_legend,
            plot_type_name="scatter_plot",
        )

    def make_line_scatter_plot(
        self,
        full_pandas_dataframe: pd.DataFrame,
        filtered_data: pd.DataFrame,
        line_plot_marker_size: int,
        y_data_variable: str,
        keys_for_output_values: List[str],
        x_data_variable: str = "Specific heating demand according to TABULA",
    ) -> None:
        """Make line scatter plot."""
        log.information("Make line scatter plot.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)

        # iterate over all scenarios
        for scenario in list(OrderedSet(list(full_pandas_dataframe[("Input", "scenario")]))):
            full_data_per_scenario = full_pandas_dataframe.loc[full_pandas_dataframe[("Input", "scenario")] == scenario]
            filtered_data_per_scenario = filtered_data.loc[filtered_data[("Input", "scenario")] == scenario]

            # get x_data_list by filtering the df according to x_data_variable and then by taking values from "value" column
            x_data_list = []
            y_data_list = []
            for output_value_key in keys_for_output_values:
                x_values = list(
                    full_data_per_scenario.loc[full_data_per_scenario[("Output", "variable")] == x_data_variable][
                        ("Output", output_value_key)
                    ].values
                )
                x_data_list.append(x_values)
                # get y values from filtered data per scenario (already filtered according to variable to check and scenario)
                y_values = list(filtered_data_per_scenario[("Output", output_value_key)].values)
                y_data_list.append(y_values)
            # get units
            x_data_unit = full_data_per_scenario.loc[full_data_per_scenario[("Output", "variable")] == x_data_variable][
                ("Output", "unit")
            ].values[0]
            y_data_unit = filtered_data_per_scenario[("Output", "unit")].values[0]

            # make scatter plot
            plt.plot(x_data_list, y_data_list, "-o", markersize=line_plot_marker_size)

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            x_axis_label=x_data_variable,
            y_axis_label=y_data_variable,
            x_axis_unit=x_data_unit,
            y_axis_unit=y_data_unit,
            show_legend=self.show_plot_legend,
            plot_type_name="line_scatter_plot",
        )

    def make_stacked_bar_plot(
        self,
        full_pandas_dataframe: pd.DataFrame,
        y1_data_variable: str,
        y2_data_variable: str,
        use_y1_as_bottom_for_y2: Optional[bool] = True,
        sort_according_to_y1_or_y2_data: Optional[str] = None,
    ) -> None:
        """Make stacked bar plot."""
        log.information("Make stacked bar plot.")

        fig, a_x = plt.subplots(figsize=self.hisim_chartbase.figsize, dpi=self.hisim_chartbase.dpi)

        # iterate over all scenarios
        y1_data_mean_value_list_for_all_scenarios = []
        y2_data_mean_value_list_for_all_scenarios = []

        for scenario in list(OrderedSet(list(full_pandas_dataframe.scenario))):
            full_data_per_scenario = full_pandas_dataframe.loc[full_pandas_dataframe["scenario"] == scenario]

            # get y1_data_list by filtering the df according to y1_data_variable and then by taking values from "value" column
            y1_data_list = list(
                full_data_per_scenario.loc[full_data_per_scenario["variable"] == y1_data_variable]["value"].values
            )
            y1_data_unit = full_data_per_scenario.loc[full_data_per_scenario["variable"] == y1_data_variable][
                "unit"
            ].values[0]
            # get y2_data_list by filtering the df according to y2_data_variable and then by taking values from "value" column
            y2_data_list = list(
                full_data_per_scenario.loc[full_data_per_scenario["variable"] == y2_data_variable]["value"].values
            )
            y2_data_unit = full_data_per_scenario.loc[full_data_per_scenario["variable"] == y2_data_variable][
                "unit"
            ].values[0]
            if y1_data_unit != y2_data_unit:
                raise ValueError("The units of y1 and y2 data variables must be the same for the stacked bar plot.")

            # if y1_data_list has more than 1 value (because more values for this scenario exist), then take mean value
            if len(y1_data_list) > 1:
                # for each scenario take the mean value
                y1_data_mean_value_per_scenario = np.mean(y1_data_list)
            elif len(y1_data_list) == 1:
                y1_data_mean_value_per_scenario = y1_data_list[0]
            else:
                raise ValueError(
                    "The y1_data_list is empty. Probably the full dataframe did not contain the y1_data_variable in the column variable."
                )
            # if y2_data_list has more than 1 value (because more values for this scenario exist), then take mean value
            if len(y2_data_list) > 1:
                # for each scenario take the mean value
                y2_data_mean_value_per_scenario = np.mean(y2_data_list)
            elif len(y2_data_list) == 1:
                y2_data_mean_value_per_scenario = y2_data_list[0]
            else:
                raise ValueError(
                    "The y2_data_list is empty. Probably the full dataframe did not contain the y2_data_variable in the column variable."
                )

            # append to y1_data_mean_value_list
            y1_data_mean_value_list_for_all_scenarios.append(y1_data_mean_value_per_scenario)

            # append to y2_data_mean_value_list
            y2_data_mean_value_list_for_all_scenarios.append(y2_data_mean_value_per_scenario)

        x_data = np.arange(0, len(y1_data_mean_value_list_for_all_scenarios) * 2, step=2)
        # x_data = list(OrderedSet(list(full_pandas_dataframe.scenario)))

        # sort values if demanded
        if sort_according_to_y1_or_y2_data == "y1":
            (
                y1_data_mean_value_list_for_all_scenarios_sorted,
                y2_data_mean_value_list_for_all_scenarios_sorted,
            ) = self.sort_y_values_according_to_data_processing_mode(
                data_processing_mode=self.data_processing_mode,
                zip_list_one=y1_data_mean_value_list_for_all_scenarios,
                zip_list_two=y2_data_mean_value_list_for_all_scenarios,
            )

        elif sort_according_to_y1_or_y2_data == "y2":
            (
                y2_data_mean_value_list_for_all_scenarios_sorted,
                y1_data_mean_value_list_for_all_scenarios_sorted,
            ) = self.sort_y_values_according_to_data_processing_mode(
                data_processing_mode=self.data_processing_mode,
                zip_list_one=y2_data_mean_value_list_for_all_scenarios,
                zip_list_two=y1_data_mean_value_list_for_all_scenarios,
            )
        else:
            y1_data_mean_value_list_for_all_scenarios_sorted = y1_data_mean_value_list_for_all_scenarios
            y2_data_mean_value_list_for_all_scenarios_sorted = y2_data_mean_value_list_for_all_scenarios

        a_x.bar(x_data, y1_data_mean_value_list_for_all_scenarios_sorted, color="r")
        if use_y1_as_bottom_for_y2 is True:
            a_x.bar(
                x_data,
                y2_data_mean_value_list_for_all_scenarios_sorted,
                bottom=y1_data_mean_value_list_for_all_scenarios_sorted,
                color="b",
            )
        else:
            a_x.bar(x_data, y2_data_mean_value_list_for_all_scenarios_sorted, color="b")

        self.set_ticks_labels_legend_and_save_fig(
            fig=fig,
            a_x=a_x,
            show_legend=True,
            plot_type_name="stacked_bar_plot",
            y_axis_unit=y1_data_unit,
            x_axis_label=str(full_pandas_dataframe.year.values[0]),
            show_x_ticks=False,
            legend_labels=[y1_data_variable, y2_data_variable],
        )

    def get_scatter_marker_size(self, data_length: int) -> int:
        """Get scatter marker size."""
        if data_length < 10:
            scatter_plot_marker_size = 20
        elif 10 < data_length < 50:
            scatter_plot_marker_size = 16
        elif 50 < data_length < 100:
            scatter_plot_marker_size = 8
        elif 100 < data_length < 300:
            scatter_plot_marker_size = 6
        elif 300 < data_length < 500:
            scatter_plot_marker_size = 4
        elif 500 < data_length < 1000:
            scatter_plot_marker_size = 2
        else:
            scatter_plot_marker_size = 1
        return scatter_plot_marker_size

    def set_axis_scale(self, a_x: Any, x_or_y: Any, unit: Any) -> Tuple[Any, str, Any]:
        """Get axis and unit and scale it properly."""

        if x_or_y == "x":
            tick_values = a_x.get_xticks()
        elif x_or_y == "y":
            tick_values = a_x.get_yticks()
        else:
            raise ValueError("x_or_y must be either 'x' or 'y'")

        max_ticks = max(tick_values)
        min_ticks = min(tick_values)

        max_scale = max(abs(max_ticks), abs(min_ticks))

        new_tick_values = tick_values
        scale = ""

        if unit not in ["-", "%", "m2"]:
            if max_scale >= 1e12:
                new_tick_values = tick_values * 1e-12
                scale = "T"
            elif 1e9 <= max_scale < 1e12:
                new_tick_values = tick_values * 1e-9
                scale = "G"
            elif 1e6 <= max_scale < 1e9:
                new_tick_values = tick_values * 1e-6
                scale = "M"
            elif 1e3 <= max_scale < 1e6:
                new_tick_values = tick_values * 1e-3
                scale = "k"
            elif -1e3 <= max_scale < 1e3:
                new_tick_values = tick_values
                scale = ""

        tick_locations = tick_values
        tick_labels = np.round(new_tick_values, 1)
        unit = f"{scale}{unit}"

        # if k already in unit, remove k and replace with "M"
        if unit in ["kkWh", "kkW"]:
            unit = "M" + unit[2:]
        elif unit in ["kkg", "kkg/s"]:
            unit = "t" + unit[3:]

        return tick_labels, unit, tick_locations

    def set_ticks_labels_legend_and_save_fig(
        self,
        fig: Any,
        a_x: Any,
        show_legend: bool,
        plot_type_name: str,
        y_axis_label: str = "",
        y_axis_unit: str = "",
        x_axis_label: str = "",
        x_axis_unit: str = "",
        legend_labels: Optional[Any] = None,
        x_ticks: Any = None,
        x_tick_labels: Any = None,
        show_x_ticks: bool = True,
        rotate_x_ticks: bool = False,
    ) -> None:
        """Set ticks, labels and legend for plot and save."""

        # y-ticks
        y_tick_labels, y_axis_unit, y_tick_locations = self.set_axis_scale(a_x=a_x, x_or_y="y", unit=y_axis_unit)
        plt.yticks(
            ticks=y_tick_locations, labels=y_tick_labels, fontsize=self.hisim_chartbase.fontsize_ticks,
        )

        # y-label
        if y_axis_label != "" and y_axis_unit != "":
            plt.ylabel(
                ylabel=f"{y_axis_label} \n [{y_axis_unit}]", fontsize=self.hisim_chartbase.fontsize_label,
            )
        elif y_axis_label != "" and y_axis_unit == "":
            plt.ylabel(
                ylabel=f"{y_axis_label}", fontsize=self.hisim_chartbase.fontsize_label,
            )
        else:
            plt.ylabel(
                ylabel=f"[{y_axis_unit}]", fontsize=self.hisim_chartbase.fontsize_label,
            )

        # x-label
        if x_axis_label != "" and x_axis_unit != "":
            plt.xlabel(
                xlabel=f"{x_axis_label} \n [{x_axis_unit}]", fontsize=self.hisim_chartbase.fontsize_label,
            )
        elif x_axis_label != "" and x_axis_unit == "":
            plt.xlabel(
                xlabel=f"{x_axis_label}", fontsize=self.hisim_chartbase.fontsize_label,
            )
        elif x_axis_label == "" and x_axis_unit != "":
            plt.xlabel(
                xlabel=f"[{x_axis_unit}]", fontsize=self.hisim_chartbase.fontsize_label,
            )
        else:
            pass

        # x-ticks
        if rotate_x_ticks:
            if x_ticks is None and x_tick_labels is None:
                a_x.tick_params(axis="x", labelrotation=45)
            else:
                a_x.set_xticks(x_ticks, x_tick_labels, rotation=45, ha="right", rotation_mode="anchor")

        if show_x_ticks:
            pass
        else:
            a_x.xaxis.set_tick_params(labelbottom=False)
            a_x.set_xticks([])

        plt.tick_params(labelsize=self.hisim_chartbase.fontsize_ticks)

        # legend
        if show_legend:
            if legend_labels is None:
                plt.legend(bbox_to_anchor=(1, 1), loc="upper left")
            else:
                plt.legend(legend_labels,)

        # save and close
        fig.savefig(
            os.path.join(self.plot_path_complete, f"{plot_type_name}.png"), bbox_inches="tight",
        )
        plt.close()

    def set_plot_colors_according_to_data_processing_mode(
        self, data_processing_mode: str, number_of_scenarios: int
    ) -> Tuple[List[str], Optional[str]]:
        """Set plot colors according to data processing mode."""
        # color_palette = list(mcolors.TABLEAU_COLORS.values())
        color_palette = sns.color_palette("Spectral", n_colors=number_of_scenarios)  #

        color: List[str] = []
        edgecolor: Optional[str] = None
        if data_processing_mode == ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name:
            for i in range(0, number_of_scenarios):
                color.append("blue")
        else:
            for i in range(0, number_of_scenarios):
                color.append(color_palette[i])
            edgecolor = "black"
        return color, edgecolor

    def sort_y_values_according_to_data_processing_mode(
        self, data_processing_mode: str, zip_list_one: List, zip_list_two: List
    ) -> Tuple[List, List]:
        """Decide whether to sort y values or not."""
        # if all data is processed and no scenario is chosen, the y values for plots should be sorted
        if data_processing_mode == ResultDataProcessingModeEnum.PROCESS_ALL_DATA.name:
            sorted_zip_lists = sorted(zip(zip_list_one, zip_list_two), reverse=True)
            list_one_sorted = [y1 for y1, y2 in sorted_zip_lists]
            list_two_sorted = [y2 for y1, y2 in sorted_zip_lists]
            return list_one_sorted, list_two_sorted
        # otherwise the order of the scenarios should be maintained
        return zip_list_one, zip_list_two

    # def make_sankey_plot_for_pyam_dataframe(
    #     self,
    #     pyam_dataframe: pyam.IamDataFrame,
    #     filter_model: Optional[str],
    #     filter_scenario: Optional[str],
    #     filter_variables: Optional[str],
    #     filter_region: Optional[str],
    #     filter_unit: Optional[str],
    #     filter_year: Optional[str],
    # ) -> None:
    #     """Make sankey plot."""
    #     log.information("Make sankey plot.")

    #     filtered_data = self.filter_pyam_dataframe(
    #         pyam_dataframe=pyam_dataframe,
    #         filter_model=filter_model,
    #         filter_scenario=filter_scenario,
    #         filter_region=filter_region,
    #         filter_variables=filter_variables,
    #         filter_unit=filter_unit,
    #         filter_year=filter_year,
    #     )

    #     sankey_mapping = {
    #         "ElectrcityGridBaseLoad|Electricity|ElectricityOutput": (
    #             "PV",
    #             "Occupancy",
    #         ),
    #         "PVSystemw-|Electricity|ElectricityOutput": ("PV", "Grid"),
    #         "Occupancy|Electricity|ElectricityOutput": ("Grid", "Occupancy"),
    #     }
    #     fig = filtered_data.plot.sankey(mapping=sankey_mapping)

    #     # save figure as html first
    #     plotly.offline.plot(
    #         fig,
    #         filename=os.path.join(self.plot_path_complete, "sankey_plot.html"),
    #         auto_open=False,
    #     )

    #     # convert html file to png
    #     hti = Html2Image()
    #     with open(
    #         os.path.join(self.plot_path_complete, "sankey_plot.html"), encoding="utf8",
    #     ) as file:
    #         hti.screenshot(
    #             file.read(), save_as="sankey_plot.png",
    #         )

    #     # change directory of sankey output file
    #     try:
    #         os.rename(
    #             "sankey_plot.png",
    #             os.path.join(self.plot_path_complete, "sankey_plot.png"),
    #         )
    #     except Exception as exc:
    #         raise Exception("Cannot save current sankey. Try again.") from exc



""" Main module for HiSim: Starts the Simulator. """
# clean
import warnings
import importlib
import os
import sys
from datetime import datetime
from typing import Optional
from pathlib import Path

from dotenv import load_dotenv

import hisim.simulator as sim
from hisim import log
from hisim.simulationparameters import SimulationParameters

load_dotenv()


def main(
    path_to_module: str,
    my_simulation_parameters: Optional[SimulationParameters] = None,
    my_module_config: Optional[str] = None,
) -> None:
    """Core function."""
    # filter warnings due to pvlib, pvlib generates warnings during simulation within pvlib package
    warnings.filterwarnings("ignore")
    # before starting, delete old logging files if path and logging files exist
    logging_default_path = log.LOGGING_DEFAULT_PATH
    if os.path.exists(logging_default_path) and os.listdir(logging_default_path) != []:
        for file in os.listdir(logging_default_path):
            if os.path.exists(os.path.join(logging_default_path, file)):
                try:
                    os.remove(os.path.join(logging_default_path, file))
                except Exception:
                    log.information(
                        "Logging default file could not be removed. This can occur when more than one simulation run simultaneously."
                    )

    function_in_module = "setup_function"
    log.information("#################################")
    log.information("starting simulation of " + path_to_module)
    starttime = datetime.now()
    starting_date_time_str = starttime.strftime("%d-%b-%Y %H:%M:%S")
    log.information("Start @ " + starting_date_time_str + " ")
    log.profile(path_to_module + " " + function_in_module + "Start @ " + starting_date_time_str)
    log.information("#################################")
    normalized_path = os.path.normpath(path_to_module)
    path_in_list = normalized_path.split(os.sep)
    module_filename = path_in_list[-1]
    if len(path_in_list) >= 1:
        if Path(path_to_module).is_absolute():
            path_to_be_added = os.path.join(os.path.abspath(os.sep), *path_in_list[:-1])
        else:
            path_to_be_added = os.path.join(os.getcwd(), *path_in_list[:-1])
        if os.path.isdir(path_to_be_added):
            #  Add current path to PYTHONPATH
            sys.path.append(path_to_be_added)
        else:
            raise ValueError(
                f"Directory location of module location is nonexistent!\nDirectory entered: {path_to_be_added}"
            )
    suffix = module_filename[-3:]
    if suffix != ".py":
        module_full_filename = f"{module_filename}.py"
    else:
        module_full_filename = module_filename
        module_filename = module_filename[:-3]
    filepath = os.path.join(path_to_be_added, module_full_filename)
    if os.path.isfile(filepath):
        # Get setup function to executable
        targetmodule = importlib.import_module(module_filename)
    else:
        raise ValueError(f"Python script {module_filename}.py could not be found")

    # Create a Simulator object based on setup function
    my_sim: sim.Simulator = sim.Simulator(
        module_directory=path_to_be_added,
        module_filename=module_filename,
        setup_function=function_in_module,
        my_simulation_parameters=my_simulation_parameters,
        my_module_config=my_module_config,
    )

    # Build method
    model_init_method = getattr(targetmodule, function_in_module)

    # Pass setup function to simulator
    model_init_method(my_sim, my_simulation_parameters)

    # Perform simulation throughout the defined timeline
    my_sim.run_all_timesteps()

    log.information("#################################")
    endtime = datetime.now()
    starting_date_time_str = endtime.strftime("%d-%b-%Y %H:%M:%S")
    log.information("finished @ " + starting_date_time_str)
    log.profile("finished @ " + starting_date_time_str)
    log.profile("duration: " + str((endtime - starttime).total_seconds()))
    log.information("#################################")
    log.information("")

    # At the end put new logging files into result directory
    try:
        my_sim.put_log_files_into_result_path()
    # sometimes when running many simulations at once this leads to errors, so ignore
    except Exception:
        pass


if __name__ == "__main__":
    if len(sys.argv) < 2:
        log.information("HiSim needs at least one argument.")
        sys.exit(1)
    FILE_NAME = sys.argv[1]
    FUNCTION_NAME = "setup_function"
    if len(sys.argv) == 2:
        log.information("calling " + FUNCTION_NAME + " from " + FILE_NAME)
        main(path_to_module=FILE_NAME)
    if len(sys.argv) == 3:
        MODULE_CONFIG = sys.argv[2]
        log.information("calling " + FUNCTION_NAME + " from " + FILE_NAME + " with module config " + MODULE_CONFIG)
        main(
            path_to_module=FILE_NAME,
            my_module_config=MODULE_CONFIG,
        )



""" Class for the simulation repository. """
# clean
from typing import Any, Dict
from threading import Lock
import enum
from hisim import loadtypes as lt


# https://refactoring.guru/design-patterns/singleton/python/example#example-1


class SingletonMeta(type):

    """A class for a thread-safe implementation of Singleton."""

    _instances: Dict[Any, Any] = {}

    _lock: Lock = Lock()
    # We now have a lock object that will be used to synchronize threads during first access to the Singleton.

    def __call__(cls, *args, **kwargs):
        """Possible changes to the value of the `__init__` argument do not affect the returned instance."""
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.
        with cls._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SingletonSimRepository(metaclass=SingletonMeta):

    """Class for exchanging information across all components."""

    def __init__(self) -> None:
        """Initializes the SimRepository."""
        self.my_dict: Dict[Any, Any] = {}
        self.my_dynamic_dict: Dict[lt.ComponentType, Dict[int, Any]] = {elem: {} for elem in lt.ComponentType}

    def set_entry(self, key: Any, entry: Any) -> None:
        """Sets an entry in the SimRepository."""
        self.my_dict[key] = entry

    def get_entry(self, key: Any) -> Any:
        """Gets an entry from the SimRepository."""
        return self.my_dict[key]

    def exist_entry(self, key: Any) -> bool:
        """Checks if an entry exists."""
        if key in self.my_dict:
            return True
        return False

    def delete_entry(self, key: Any) -> None:
        """Deletes an existing entry."""
        self.my_dict.pop(key)

    def set_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int, entry: Any) -> None:
        """Sets a dynamic entry."""
        self.my_dynamic_dict[component_type][source_weight] = entry

    def get_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int) -> Any:
        """Gets a dynmaic entry."""
        component = self.my_dynamic_dict.get(component_type, None)
        if component is None:
            return None
        value = component.get(source_weight, None)
        return value

    def get_dynamic_component_weights(self, component_type: lt.ComponentType) -> list:
        """Gets weights for dynamic components."""
        return list(self.my_dynamic_dict[component_type].keys())

    def delete_dynamic_entry(self, component_type: lt.ComponentType, source_weight: int) -> Any:
        """Deletes a dynamic component entry."""
        self.my_dynamic_dict[component_type].pop(source_weight)

    def clear(self):
        """Clears all dictionaries at the end of the simulation to enable garbage collection and reduce memory consumption."""
        self.my_dict.clear()
        del self.my_dict
        self.my_dynamic_dict.clear()
        del self.my_dynamic_dict


class SingletonDictKeyEnum(enum.Enum):

    """Class for setting dictionary keys in the singleton sim repository."""

    NUMBEROFAPARTMENTS = 1
    WATERMASSFLOWRATEOFHEATGENERATOR = 2
    MAXTHERMALBUILDINGDEMAND = 3
    SETHEATINGTEMPERATUREFORWATERSTORAGE = 4
    SETCOOLINGTEMPERATUREFORWATERSTORAGE = 5
    LOCATION = 6
    RESULT_SCENARIO_NAME = 7
    THERMALTRANSMISSIONCOEFFICIENTGLAZING = 8
    THERMALTRANSMISSIONSURFACEINDOORAIR = 9
    THERMALTRANSMISSIONCOEFFICIENTOPAQUEEM = 10
    THERMALTRANSMISSIONCOEFFICIENTOPAQUEMS = 11
    THERMALTRANSMISSIONCOEFFICIENTVENTILLATION = 12
    THERMALCAPACITYENVELOPE = 13
    PREDICTIVE = 14
    PREDICTIONHORIZON = 15
    PVINCLUDED = 16
    PVPEAKPOWER = 17
    SMARTDEVICESINCLUDED = 18
    BATTERYINCLUDED = 19
    MPCBATTERYCAPACITY = 20
    COPCOEFFICIENTHEATING = 21
    EERCOEFFICIENTCOOLING = 22
    WEATHERTEMPERATUREOUTSIDEYEARLYFORECAST = 23
    HEATFLUXTHERMALMASSNODEFORECAST = 24
    HEATFLUXSURFACENODEFORECAST = 25
    HEATFLUXINDOORAIRNODEFORECAST = 26
    PVFORECASTYEARLY = 28
    MAXIMUMBATTERYCAPACITY = 29
    MINIMUMBATTERYCAPACITY = 30
    MAXIMALCHARGINGPOWER = 31
    MAXIMALDISCHARGINGPOWER = 32
    BATTERYEFFICIENCY = 33
    INVERTEREFFICIENCY = 34
    PRICEPURCHASEFORECAST24H = 35
    PRICEINJECTIONFORECAST24H = 36
    WEATHERALTITUDEYEARLYFORECAST = 37
    WEATHERDIFFUSEHORIZONTALIRRADIANCEYEARLYFORECAST = 38
    WEATHERDIRECTNORMALIRRADIANCEYEARLYFORECAST = 39
    WEATHERDIRECTNORMALIRRADIANCEEXTRAYEARLYFORECAST = 40
    WEATHERGLOBALHORIZONTALIRRADIANCEYEARLYFORECAST = 41
    WEATHERAZIMUTHYEARLYFORECAST = 42
    WEATHERAPPARENTZENITHYEARLYFORECAST = 43
    HEATINGBYRESIDENTSYEARLYFORECAST = 44
    WEATHERWINDSPEEDYEARLYFORECAST = 45
    WEATHERPRESSUREYEARLYFORECAST = 46



""" Charts for a single day. """
# clean
import os
from typing import Any
import matplotlib.pyplot as plt
from matplotlib.axis import Axis
from matplotlib.dates import DateFormatter

from hisim.postprocessing.chartbase import Chart, ChartFontsAndSize
from hisim.postprocessing.report_image_entries import ReportImageEntry


class ChartSingleDay(Chart, ChartFontsAndSize):

    """For making visualisations for a single day."""

    def __init__(
        self,
        output: Any,
        component_name: str,
        units: Any,
        directory_path: str,
        time_correction_factor: float,
        output_description: str,
        data: Any,
        day: Any = 0,
        month: Any = 0,
        output2: Any = None,
        figure_format: Any = None,
    ):
        """Initializes the class."""
        if output2 is not None:
            super().__init__(
                output=output,
                component_name=component_name,
                chart_type="days",
                units=units,
                directory_path=directory_path,
                time_correction_factor=time_correction_factor,
                output_description=output_description,
                output2=output2,
                figure_format=figure_format,
            )
        else:
            super().__init__(
                output=output,
                component_name=component_name,
                chart_type="days",
                units=units,
                directory_path=directory_path,
                time_correction_factor=time_correction_factor,
                output_description=output_description,
                figure_format=figure_format,
            )

        self.month = month
        self.day = day
        self.data = data
        self.plot_title: str
        self.filename = (
            f"{self.type.lower()}_m"
            f"{self.month}_d{self.day}{self.figure_format}"
        )

        self.filepath = os.path.join(self.directory_path, self.filename)
        self.filepath2 = os.path.join(self.component_output_folder_path, self.filename)

    def get_day_data(self):
        """Extracts data for a single day."""
        firstindex = (self.month * 30 + self.day) * 24 * int(1 / self.time_correction_factor)
        lastindex = firstindex + 24 * int(1 / self.time_correction_factor)
        day_number = self.day + 1
        if day_number == 1:
            ordinal = "st"
        elif day_number == 2:
            ordinal = "nd"
        elif day_number == 3:
            ordinal = "rd"
        else:
            ordinal = "th"
        date = f"{self.label_months_lowercase[self.month]} {day_number}{ordinal}"
        self.plot_title = f"{self.title} {date}"

        if abs(lastindex - firstindex) < len(self.data):
            data = self.data[firstindex:lastindex]
            data_index = data.index[firstindex:lastindex]
            data.index = data_index
            return data
        return self.data

    def close(self):
        """Closes a chart and saves."""

        plt.savefig(self.filepath2)
        plt.close()

    def plot(self, close: Any) -> ReportImageEntry:
        """Plots a chart."""
        single_day_data = self.get_day_data()
        plt.rcParams["font.size"] = "30"
        plt.rcParams["agg.path.chunksize"] = 10000
        _fig, a_x = plt.subplots(figsize=self.figsize, dpi=self.dpi)
        plt.xticks(fontsize=self.fontsize_ticks)
        plt.yticks(fontsize=self.fontsize_ticks)

        single_day_data, self.units = self.rescale_y_axis(y_values=single_day_data, units=self.units)
        plt.title(self.title, fontsize=self.fontsize_title)
        plt.plot(
            single_day_data.index, single_day_data, color="green", linewidth=1.0, label=self.property,
        )
        plt.grid(True)
        plt.xlabel("Time [hours]", fontsize=self.fontsize_label)
        plt.ylabel(f"[{self.units}]", fontsize=self.fontsize_label)
        plt.tight_layout()
        Axis.set_major_formatter(a_x.xaxis, DateFormatter("%H:%M"))
        if close:
            self.close()
        return ReportImageEntry(
            category=None,
            output_description=self.output_description,
            component_output_folder_path=self.component_output_folder_path,
            file_path=self.filepath2,
            unit=self.units,
            component_name=self.component_name,
            output_type=self.output_type,
        )



""" Battery implementation built upon the bslib library. It contains a Battery Class together with its Configuration and State. """

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
from typing import List, Tuple
from dataclasses import dataclass
from bslib import bslib as bsl
from dataclasses_json import dataclass_json

import pandas as pd

# Import modules from HiSim
from hisim.component import (
    Component,
    ComponentInput,
    ComponentOutput,
    SingleTimeStepValues,
    ConfigBase,
    OpexCostDataClass,
    DisplayConfig,
    CapexCostDataClass
)
from hisim.loadtypes import LoadTypes, Units, InandOutputType, ComponentType
from hisim.simulationparameters import SimulationParameters
from hisim import log
from hisim.postprocessing.kpi_computation.kpi_structure import KpiTagEnumClass, KpiEntry

__authors__ = "Tjarko Tjaden, Hauke Hoops, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = "..."
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Tjarko Tjaden"
__email__ = "tjarko.tjaden@hs-emden-leer.de"
__status__ = "development"


@dataclass_json
@dataclass
class BatteryConfig(ConfigBase):
    """Battery Configuration."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return Battery.get_full_classname()

    #: building_name in which component is
    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    #: name of battery to search in database (bslib)
    system_id: str
    #: charging and discharging power in Watt
    custom_pv_inverter_power_generic_in_watt: float
    #: battery capacity in in kWh
    custom_battery_capacity_generic_in_kilowatt_hour: float
    #: amount of energy used to charge the car battery
    charge_in_kwh: float
    #: amount of energy discharged from the battery
    discharge_in_kwh: float
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    #: lifetime of battery in years
    lifetime: float
    #: lifetime of battery in full cycles
    lifetime_in_cycles: float
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float

    @classmethod
    def get_default_config(cls, building_name: str = "BUI1", name: str = "Battery") -> "BatteryConfig":
        """Returns default configuration of battery."""
        custom_battery_capacity_generic_in_kilowatt_hour = (
            10  # size/capacity of battery should be approx. the same as default pv power
        )
        config = BatteryConfig(
            building_name=building_name,
            name=name,
            # https://www.energieinstitut.at/die-richtige-groesse-von-batteriespeichern/
            custom_battery_capacity_generic_in_kilowatt_hour=custom_battery_capacity_generic_in_kilowatt_hour,
            custom_pv_inverter_power_generic_in_watt=10 * 0.5 * 1e3,  # c-rate is 0.5C (0.5/h) here
            source_weight=1,
            system_id="SG1",
            charge_in_kwh=0,
            discharge_in_kwh=0,
            co2_footprint=custom_battery_capacity_generic_in_kilowatt_hour
            * 130.7,  # value from emission_factros_and_costs_devices.csv
            cost=custom_battery_capacity_generic_in_kilowatt_hour
            * 535.81,  # value from emission_factros_and_costs_devices.csv
            lifetime=10,  # estimated value , source: https://pv-held.de/wie-lange-haelt-batteriespeicher-photovoltaik/
            lifetime_in_cycles=5e3,  # estimated value , source: https://pv-held.de/wie-lange-haelt-batteriespeicher-photovoltaik/
            maintenance_cost_as_percentage_of_investment=0.02,  # SOURCE: https://solarenergie.de/stromspeicher/preise
        )
        return config

    @classmethod
    def get_scaled_battery(
        cls, total_pv_power_in_watt_peak: float, building_name: str = "BUI1", name: str = "Battery"
    ) -> "BatteryConfig":
        """Returns scaled configuration of battery according to pv power."""
        custom_battery_capacity_generic_in_kilowatt_hour = (
            total_pv_power_in_watt_peak * 1e-3
        )  # size/capacity of battery should be approx. the same as default pv power
        c_rate = 0.5  # 0.5C corresponds to 0.5/h for fully charging or discharging
        config = BatteryConfig(
            building_name=building_name,
            name=name,
            # https://www.energieinstitut.at/die-richtige-groesse-von-batteriespeichern/
            custom_battery_capacity_generic_in_kilowatt_hour=custom_battery_capacity_generic_in_kilowatt_hour,
            custom_pv_inverter_power_generic_in_watt=custom_battery_capacity_generic_in_kilowatt_hour * c_rate * 1e3,
            source_weight=1,
            system_id="SG1",
            charge_in_kwh=0,
            discharge_in_kwh=0,
            co2_footprint=custom_battery_capacity_generic_in_kilowatt_hour
            * 130.7,  # value from emission_factros_and_costs_devices.csv
            cost=custom_battery_capacity_generic_in_kilowatt_hour
            * 535.81,  # value from emission_factros_and_costs_devices.csv
            lifetime=10,  # todo set correct values
            lifetime_in_cycles=5e3,  # todo set correct values
            maintenance_cost_as_percentage_of_investment=0.02,  # SOURCE: https://solarenergie.de/stromspeicher/preise
        )

        return config


class Battery(Component):
    """Battery class.

    Simulate state of charge and realized power of a ac coupled battery
    storage system with the bslib library. Relevant simulation parameters
    are loaded within the init for a specific or generic battery type.

    Components to connect to:
    (1) Energy Management System
    """

    # Inputs
    LoadingPowerInput = "LoadingPowerInput"  # W

    # Outputs
    AcBatteryPowerUsed = "AcBatteryPowerUsed"  # W
    DcBatteryPowerUsed = "DcBatteryPowerUsed"  # W
    StateOfCharge = "StateOfCharge"  # [0..1]

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: BatteryConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Loads the parameters of the specified battery storage."""
        self.battery_config = config

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.source_weight = self.battery_config.source_weight

        self.system_id = self.battery_config.system_id

        self.custom_pv_inverter_power_generic_in_watt = self.battery_config.custom_pv_inverter_power_generic_in_watt

        self.custom_battery_capacity_generic_in_kilowatt_hour = (
            self.battery_config.custom_battery_capacity_generic_in_kilowatt_hour
        )

        # Component has states
        self.state = BatteryState()
        self.previous_state = self.state.clone()

        # Load battery object with parameters from bslib database
        self.ac_coupled_battery_object = bsl.ACBatMod(
            system_id=self.system_id,
            p_inv_custom=self.custom_pv_inverter_power_generic_in_watt,
            e_bat_custom=self.custom_battery_capacity_generic_in_kilowatt_hour,
        )

        # Define component inputs
        self.loading_power_input_channel: ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.LoadingPowerInput,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            mandatory=True,
        )

        # Define component outputs
        self.ac_battery_power_channel: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.AcBatteryPowerUsed,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            postprocessing_flag=[
                InandOutputType.CHARGE_DISCHARGE,
                ComponentType.BATTERY,
            ],
            output_description=f"here a description for {self.AcBatteryPowerUsed} will follow.",
        )

        self.dc_battery_power_channel: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.DcBatteryPowerUsed,
            load_type=LoadTypes.ELECTRICITY,
            unit=Units.WATT,
            output_description=f"here a description for {self.DcBatteryPowerUsed} will follow.",
        )

        self.state_of_charge_channel: ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.StateOfCharge,
            load_type=LoadTypes.ANY,
            unit=Units.ANY,
            postprocessing_flag=[InandOutputType.STORAGE_CONTENT],
            output_description=f"here a description for {self.StateOfCharge} will follow.",
        )

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        # Parameters
        time_increment_in_seconds = self.my_simulation_parameters.seconds_per_timestep

        # Load input values
        set_point_for_ac_battery_power_in_watt = stsv.get_input_value(self.loading_power_input_channel)
        state_of_charge = self.state.state_of_charge

        # Simulate on timestep
        results = self.ac_coupled_battery_object.simulate(
            p_load=set_point_for_ac_battery_power_in_watt,
            soc=state_of_charge,
            dt=time_increment_in_seconds,
        )
        # The bslib simulation returns how much of loading power input was actually used for charging and discharging and the resulting state of charge
        ac_battery_power_used_for_charging_or_discharging_in_watt = results[0]
        dc_battery_power_used_for_charging_or_discharging_in_watt = results[1]
        state_of_charge = results[2]

        # write values for output time series
        stsv.set_output_value(self.ac_battery_power_channel, ac_battery_power_used_for_charging_or_discharging_in_watt)
        stsv.set_output_value(self.dc_battery_power_channel, dc_battery_power_used_for_charging_or_discharging_in_watt)
        stsv.set_output_value(self.state_of_charge_channel, state_of_charge)

        # write values to state
        self.state.state_of_charge = state_of_charge

    def write_to_report(self) -> List[str]:
        """Write to report."""
        return self.battery_config.get_string_dict()

    @staticmethod
    def get_battery_aging_information(
        config: BatteryConfig
    ) -> Tuple[float, float]:
        """Calculate battery aging.

        This is used to calculate investment costs for battery per simulated period.
        Battery aging is ROUGHLY approximated by costs for each virtual charging cycle used in simulated period
        (costs_per_cycle = investment / lifetime_in_cycles).
        """
        # Todo: Think about better approximation for costs of battery aging

        virtual_number_of_full_charge_cycles = (
            config.charge_in_kwh / config.custom_battery_capacity_generic_in_kilowatt_hour
        )
        # virtual_number_of_full_discharge_cycles = self.battery_config.discharge_in_kwh / self.battery_config.custom_battery_capacity_generic_in_kilowatt_hour

        return virtual_number_of_full_charge_cycles, config.lifetime_in_cycles

    @staticmethod
    def get_cost_capex(config: BatteryConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        # Todo: think about livetime in cycles not in years
        (virtual_number_of_full_charge_cycles, lifetime_in_cycles) = Battery.get_battery_aging_information(config=config)
        if lifetime_in_cycles > 0:
            capex_per_simulated_period = (config.cost / lifetime_in_cycles) * (virtual_number_of_full_charge_cycles)
            device_co2_footprint_per_simulated_period = (config.co2_footprint / lifetime_in_cycles) * (
                virtual_number_of_full_charge_cycles
            )
        else:
            log.warning(
                "Capex calculation not valid. Check lifetime_in_cycles in Configuration of Battery."
            )

        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost,
            device_co2_footprint_in_kg=config.co2_footprint,
            lifetime_in_years=config.lifetime,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.BATTERY
        )
        return capex_cost_data_class

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of maintenance costs."""
        battery_losses_in_kwh: float = 0.0
        for index, output in enumerate(all_outputs):
            if (
                output.postprocessing_flag is not None
                and output.component_name == self.component_name
            ):
                if InandOutputType.CHARGE_DISCHARGE in output.postprocessing_flag:
                    self.battery_config.charge_in_kwh = round(
                        postprocessing_results.iloc[:, index].clip(lower=0).sum()
                        * self.my_simulation_parameters.seconds_per_timestep
                        / 3.6e6,
                        1,
                    )
                    self.battery_config.discharge_in_kwh = round(
                        postprocessing_results.iloc[:, index].clip(upper=0).sum()
                        * self.my_simulation_parameters.seconds_per_timestep
                        / 3.6e6,
                        1,
                    ) * (-1)
                    battery_losses_in_kwh = self.battery_config.charge_in_kwh - self.battery_config.discharge_in_kwh

        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=0,
            consumption_in_kwh=battery_losses_in_kwh,
            loadtype=LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.BATTERY
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        return []


@dataclass
class BatteryState:
    """Battery state class."""

    #: state of charge of the battery
    state_of_charge: float = 0

    def clone(self):
        """Creates a copy of the Battery State."""
        return BatteryState(state_of_charge=self.state_of_charge)



# -*- coding: utf-8 -*-
# clean

""" Generic CHP controller with minimal runtime.

Heat is transfered to the drain hot water storage and either the buffer storage or the building directly.
CHP is controlled by both (i) thermal demand and (ii) electricity demand - it is only activated when both electricity and heat are needed.
"""

# Owned
import importlib
from dataclasses import dataclass
from typing import List
from dataclasses_json import dataclass_json

# Generic/Built-in
from hisim import component as cp
from hisim import utils
from hisim.component import ConfigBase
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L1CHPControllerConfig(ConfigBase):
    """CHP Controller Config."""

    building_name: str
    #: name of the device
    name: str
    #: priority of the device in hierachy: the higher the number the lower the priority
    source_weight: int
    #: type of CHP: hydrogen or gas (hydrogen than considers also SOC of hydrogen storage)
    use: LoadTypes
    #: minimal electricity demand to start operating, given in W:
    electricity_threshold: float
    #: minimal state of charge of the hydrogen storage to start operating in percent (only relevant for fuel cell):
    h2_soc_threshold: float
    #: lower set temperature of building (or buffer storage), given in °C
    t_min_heating_in_celsius: float
    #: upper set temperature of building (or buffer storage), given in °C
    t_max_heating_in_celsius: float
    #: lower set temperature of drain hot water storage, given in °C
    t_min_dhw_in_celsius: float
    #: upper set temperature of drain hot water storage, given in °C
    t_max_dhw_in_celsius: float
    # julian day of simulation year, where heating season begins
    day_of_heating_season_begin: int
    # julian day of simulation year, where heating season ends
    day_of_heating_season_end: int
    # minimal operation time of heat source
    min_operation_time_in_seconds: int
    # minimal resting time of heat source
    min_idle_time_in_seconds: int

    @staticmethod
    def get_default_config_chp(
        building_name: str = "BUI1",
    ) -> "L1CHPControllerConfig":
        """Returns default configuration for the CHP controller."""
        config = L1CHPControllerConfig(
            building_name=building_name,
            name="CHP Controller",
            source_weight=1,
            use=LoadTypes.GAS,
            electricity_threshold=300,
            h2_soc_threshold=0,
            t_min_heating_in_celsius=20.0,
            t_max_heating_in_celsius=20.5,
            t_min_dhw_in_celsius=42,
            t_max_dhw_in_celsius=60,
            day_of_heating_season_begin=270,
            day_of_heating_season_end=150,
            min_operation_time_in_seconds=3600 * 4,
            min_idle_time_in_seconds=3600 * 2,
        )
        return config

    @staticmethod
    def get_default_config_fuel_cell(
        building_name: str = "BUI1",
    ) -> "L1CHPControllerConfig":
        """Returns default configuration for the fuel cell controller."""
        config = L1CHPControllerConfig(
            building_name=building_name,
            name="Fuel Cell Controller",
            source_weight=1,
            use=LoadTypes.HYDROGEN,
            electricity_threshold=300,
            h2_soc_threshold=8.0,
            t_min_heating_in_celsius=20.0,
            t_max_heating_in_celsius=20.5,
            t_min_dhw_in_celsius=50,
            t_max_dhw_in_celsius=60,
            day_of_heating_season_begin=270,
            day_of_heating_season_end=150,
            min_operation_time_in_seconds=3600 * 4,
            min_idle_time_in_seconds=3600 * 2,
        )
        return config

    @staticmethod
    def get_default_config_chp_with_buffer(
        building_name: str = "BUI1",
    ) -> "L1CHPControllerConfig":
        """Returns default configuration for the CHP controller, when buffer storage for heating is available."""
        # minus - 1 in heating season, so that buffer heats up one day ahead, and modelling to building works.
        config = L1CHPControllerConfig(
            building_name=building_name,
            name="CHP Controller",
            source_weight=1,
            use=LoadTypes.GAS,
            electricity_threshold=300,
            h2_soc_threshold=0,
            t_min_heating_in_celsius=35.0,
            t_max_heating_in_celsius=40.0,
            t_min_dhw_in_celsius=50,
            t_max_dhw_in_celsius=60,
            day_of_heating_season_begin=270 - 1,
            day_of_heating_season_end=150,
            min_operation_time_in_seconds=3600 * 4,
            min_idle_time_in_seconds=3600 * 2,
        )
        return config

    @staticmethod
    def get_default_config_fuel_cell_with_buffer(
        building_name: str = "BUI1",
    ) -> "L1CHPControllerConfig":
        """Returns default configuration for the fuel cell controller, when buffer storage for heating is available."""
        # minus - 1 in heating season, so that buffer heats up one day ahead, and modelling to building works.
        config = L1CHPControllerConfig(
            building_name=building_name,
            name="CHP Controller",
            source_weight=1,
            use=LoadTypes.HYDROGEN,
            electricity_threshold=300,
            h2_soc_threshold=8.0,
            t_min_heating_in_celsius=31.0,
            t_max_heating_in_celsius=40.0,
            t_min_dhw_in_celsius=42,
            t_max_dhw_in_celsius=60,
            day_of_heating_season_begin=270 - 1,
            day_of_heating_season_end=150,
            min_operation_time_in_seconds=3600 * 4,
            min_idle_time_in_seconds=3600 * 2,
        )
        return config


class L1CHPControllerState:
    """Data class that saves the state of the CHP controller."""

    def __init__(
        self,
        on_off: int,
        mode: int,
        activation_time_step: int,
        deactivation_time_step: int,
    ) -> None:
        """Initializes CHP Controller state.

        :param on_off: 0 if turned off, 1 if running.
        :type on_off: int
        :param mode: 0 if water heating, 1 if heating.
        :type mode: int
        :param activation_time_step: timestep of activation (simulation time step).
        :type activation_time_step: int
        :param deactivation_time_step: timestep of deactivation
        :type deactivation_time_step: int
        """
        self.on_off: int = on_off
        self.mode: int = mode
        self.activation_time_step: int = activation_time_step
        self.deactivation_time_step: int = deactivation_time_step

    def clone(self) -> "L1CHPControllerState":
        """Copies the current instance."""
        return L1CHPControllerState(
            on_off=self.on_off,
            mode=self.mode,
            activation_time_step=self.activation_time_step,
            deactivation_time_step=self.deactivation_time_step,
        )

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def activate(self, timestep: int) -> None:
        """Activates the heat pump and remembers the time step."""
        if self.on_off == 0:
            self.activation_time_step = timestep
        self.on_off = 1

    def deactivate(self, timestep: int) -> None:
        """Deactivates the heat pump and remembers the time step."""

        if self.on_off == 1:
            self.deactivation_time_step = timestep
        self.on_off = 0


class L1CHPController(cp.Component):
    """L1 combined heat and power (CHP) or Fuel Cell Controller.

    Is activated when both Electricity and Heat are demanded. Decides if heat is transferred to
    Building or HotWaterStorage.
    When it is a fuel cell, also the SOC of the hydrogen storage is checked.

    Components to connect to:
    (1) Hot Water Storage (generic_hot_water_storage_modular)
    (2) Either buffer storage or building (generic_hot_water_storage_modular or building)
    (3) EMS controller (controller_l2_energy_management_system) -> optional if electricity should be involved in control.
    (4) Hydrogen storage (generic_hydrogen_storage) -> optional if component is fuel cell and amount of hydrogen in storage is relevant.
    """

    # Inputs
    BuildingTemperature = "BuildingTemperature"
    HotWaterStorageTemperature = "HotWaterStorageTemperature"
    ElectricityTarget = "ElectricityTarget"
    HydrogenSOC = "HydrogenSOC"

    # Outputs
    CHPControllerOnOffSignal = "CHPControllerOnOffSignal"
    CHPControllerHeatingModeSignal = "CHPControllerHeatingModeSignal"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: L1CHPControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """For initializing."""
        if not config.__class__.__name__ == L1CHPControllerConfig.__name__:
            raise ValueError("Wrong config class. Got a " + config.__class__.__name__)
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.minimum_runtime_in_timesteps = int(
            config.min_operation_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )
        self.minimum_resting_time_in_timesteps = int(
            config.min_idle_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )
        """ Initializes the class. """
        if config.day_of_heating_season_begin is None:
            raise ValueError("Day of heating season begin was None")
        if config.day_of_heating_season_end is None:
            raise ValueError("Day of heating season end was None")
        self.heating_season_begin = (
            config.day_of_heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        )
        self.heating_season_end = (
            config.day_of_heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        )
        self.state: L1CHPControllerState = L1CHPControllerState(0, 0, 0, 0)
        self.previous_state: L1CHPControllerState = self.state.clone()
        self.processed_state: L1CHPControllerState = self.state.clone()

        # Component Outputs
        self.chp_onoff_signal_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.CHPControllerOnOffSignal,
            LoadTypes.ON_OFF,
            Units.BINARY,
            output_description="On off signal from CHP controller.",
        )

        self.chp_heatingmode_signal_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.CHPControllerHeatingModeSignal,
            LoadTypes.ANY,
            Units.BINARY,
            output_description="Heating mode signal from CHP controller.",
        )

        # Component Inputs
        self.building_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BuildingTemperature,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=True,
        )
        self.dhw_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.HotWaterStorageTemperature,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=True,
        )

        self.electricity_target_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityTarget,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            mandatory=False,
        )

        self.hydrogen_soc_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.HydrogenSOC,
            LoadTypes.HYDROGEN,
            Units.PERCENT,
            mandatory=False,
        )

        self.add_default_connections(self.get_default_connections_generic_hot_water_storage_modular())
        self.add_default_connections(self.get_default_connections_from_building())
        self.add_default_connections(self.get_default_connections_from_h2_storage())

    def get_default_connections_generic_hot_water_storage_modular(self):
        """Sets default connections for the boiler."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_hot_water_storage_modular"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "HotWaterStorage")
        connections = []
        boiler_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1CHPController.HotWaterStorageTemperature,
                boiler_classname,
                component_class.TemperatureMean,
            )
        )
        return connections

    def get_default_connections_from_building(self):
        """Sets default connections for the boiler."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.building"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "Building")
        connections = []
        building_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1CHPController.BuildingTemperature,
                building_classname,
                component_class.TemperatureMeanThermalMass,
            )
        )
        return connections

    def get_default_connections_from_h2_storage(self):
        """Sets default connections for the hydrogen storage."""
        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.generic_hydrogen_storage"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "GenericHydrogenStorage")
        connections = []
        h2_storage_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1CHPController.HydrogenSOC,
                h2_storage_classname,
                component_class.HydrogenSOC,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """For double checking results."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Core Simulation function."""
        if force_convergence:
            # states are saved after each timestep, outputs after each iteration
            # outputs have to be in line with states, so if convergence is forced outputs are aligned to last known state.
            self.state = self.processed_state.clone()
        else:
            # control temperatures of boiler and building or buffer
            t_building = stsv.get_input_value(self.building_temperature_channel)
            t_dhw = stsv.get_input_value(self.dhw_temperature_channel)
            # surplus/deficit electricity threshold exceeded?
            if self.electricity_target_channel.source_output is not None:
                electricity_target = stsv.get_input_value(self.electricity_target_channel)
                electricity_threshold_ok = (
                    electricity_target <= -self.config.electricity_threshold and self.state.on_off == 0
                ) or (electricity_target <= self.config.electricity_threshold and self.state.on_off == 1)
            else:
                electricity_threshold_ok = True
            if self.hydrogen_soc_channel.source_output is not None:
                hydrogen_soc = stsv.get_input_value(self.hydrogen_soc_channel)
                hydrogen_soc_ok = hydrogen_soc >= self.config.h2_soc_threshold
            else:
                hydrogen_soc_ok = True
            self.determine_heating_mode(timestep, t_building, t_dhw)
            self.calculate_state(timestep, t_building, t_dhw, electricity_threshold_ok, hydrogen_soc_ok)
            self.processed_state = self.state.clone()
        stsv.set_output_value(self.chp_onoff_signal_channel, self.state.on_off)
        stsv.set_output_value(self.chp_heatingmode_signal_channel, self.state.mode)

    def determine_heating_mode(self, timestep: int, t_building: float, t_dhw: float) -> None:
        """Determines if hot water or building should be heated.

        The mode in the state is 0 if water heating is considered and 1 if heating is enforced.
        """
        if (
            self.heating_season_begin > timestep > self.heating_season_end
            and t_building >= self.config.t_min_heating_in_celsius - 30
        ):
            # only consider water heating in summer
            self.state.mode = 0
            return
        # calculate heating level of DHW storage and building
        soc_dhw = (t_dhw - self.config.t_min_dhw_in_celsius) / (
            self.config.t_max_dhw_in_celsius - self.config.t_min_dhw_in_celsius
        )
        soc_building = (t_building - self.config.t_min_heating_in_celsius) / (
            self.config.t_max_heating_in_celsius - self.config.t_min_heating_in_celsius
        )

        if soc_building >= soc_dhw:
            self.state.mode = 0
            return
        self.state.mode = 1

    def calculate_state(  # pylint: disable=R0911
        self,
        timestep: int,
        t_building: float,
        t_dhw: float,
        electricity_threshold_ok: bool,
        hydrogen_soc_ok: bool,
    ) -> None:
        """Calculate the CHP state and activate / deactives."""
        # return device on if minimum operation time is not fulfilled and device was on in previous state
        if self.state.on_off == 1 and self.state.activation_time_step + self.minimum_runtime_in_timesteps >= timestep:
            # mandatory on, minimum runtime not reached
            return
        if (
            self.state.on_off == 0
            and self.state.deactivation_time_step + self.minimum_resting_time_in_timesteps >= timestep
        ):
            # mandatory off, minimum resting time not reached
            return
        # deactivate when electricity is not needed:
        if not electricity_threshold_ok:
            self.state.deactivate(timestep)
            return
        # deactivate when state of charge of hydrogen storage is too low:
        if not hydrogen_soc_ok:
            self.state.deactivate(timestep)
            return
        # control according to set temperatures
        if (
            self.heating_season_begin > timestep > self.heating_season_end
            and t_building >= self.config.t_min_heating_in_celsius - 30
        ):
            # only consider water heating in summer
            if t_dhw < self.config.t_min_dhw_in_celsius:
                self.state.activate(
                    timestep
                )  # activate CHP when storage temperature is too low and electricity is needed
                return
            if t_dhw > self.config.t_max_heating_in_celsius:
                self.state.deactivate(timestep)  # deactivate CHP when storage temperature is too high
                return
        else:
            if t_building < self.config.t_min_heating_in_celsius or t_dhw < self.config.t_min_dhw_in_celsius:
                # activate heating when either dhw storage or building temperature is too low and electricity is needed
                self.state.activate(timestep)
                return
            if t_building > self.config.t_max_heating_in_celsius and t_dhw > self.config.t_max_dhw_in_celsius:
                # deactivate heating when dhw storage and building temperature is too high
                self.state.deactivate(timestep)
                return

    def write_to_report(self) -> List[str]:
        """Writes the information of the current component to the report."""
        return self.config.get_string_dict()



"""Modular household config module."""

# clean

from typing import Optional
import json
from dataclasses import dataclass

from hisim.modular_household.interface_configs import archetype_config, system_config
from hisim import log
from hisim.system_setup_configuration import SystemSetupConfigBase


@dataclass
class ModularHouseholdConfig(SystemSetupConfigBase):

    """Modular Household Config class."""

    # configuration of the technological equipment of the household
    system_config_: Optional[system_config.SystemConfig] = None
    # configuration of the framework of the household (climate, house type, mobility behaviour, heating system, etc. )
    archetype_config_: Optional[archetype_config.ArcheTypeConfigModular] = None

    @classmethod
    def get_default(cls):
        """Get default ModularHouseholdConfig."""
        system_config_ = system_config.SystemConfig()
        archetype_config_ = archetype_config.ArcheTypeConfigModular()
        household_config = ModularHouseholdConfig(system_config_=system_config_, archetype_config_=archetype_config_)
        return household_config


def write_config(config: ModularHouseholdConfig) -> None:
    """Writes config."""
    with open("modular_example_config.json", "w", encoding="utf-8") as file:
        file.write(config.to_json())  # type: ignore


def read_in_configs(pathname: str) -> ModularHouseholdConfig:
    """Reads in ModularHouseholdConfig file and loads default if file cannot be found."""
    try:
        with open(pathname, encoding="utf8") as config_file:
            household_config_dict = json.load(config_file)  # type: ignore
            household_config: ModularHouseholdConfig = ModularHouseholdConfig.from_dict(
                household_config_dict.get("system_setup_config")
            )  # type: ignore
        log.information(f"Read modular household config from {pathname}")
        if (household_config.system_config_ is None) and (household_config.archetype_config_ is None):
            raise ValueError()
    except Exception:
        household_config = ModularHouseholdConfig()
        log.warning(f"Could not read the modular household config from '{pathname}'. Using a default config instead.")

    # set default configs
    if household_config.system_config_ is None:
        household_config.system_config_ = system_config.SystemConfig()
    if household_config.archetype_config_ is None:
        household_config.archetype_config_ = archetype_config.ArcheTypeConfigModular()

    return household_config



""" Data Transfer Object to get all the result data to the post processing. """

# clean
from typing import Any, List, Dict
from hisim import log
from hisim.component_wrapper import ComponentWrapper
from hisim.simulationparameters import SimulationParameters


class PostProcessingDataTransfer:  # noqa: too-few-public-methods

    """Data class for transfering the result data to this class."""

    def __init__(  # pylint: disable=dangerous-default-value
        self,
        results: Any,
        all_outputs: Any,
        simulation_parameters: SimulationParameters,
        wrapped_components: List[ComponentWrapper],
        mode: Any,
        setup_function: Any,
        module_filename: Any,
        my_module_config: Any,
        execution_time: Any,
        results_monthly: Any,
        results_hourly: Any,
        results_cumulative: Any,
        results_daily: Any,
        kpi_collection_dict: Dict = {},
    ) -> None:
        """Initializes the values."""
        # Johanna Ganglbauer: time correction factor is applied in postprocessing to sum over power values and convert them to energy
        self.time_correction_factor = simulation_parameters.seconds_per_timestep / 3600
        self.results = results
        self.all_outputs = all_outputs
        self.simulation_parameters = simulation_parameters
        self.wrapped_components: List[ComponentWrapper] = wrapped_components
        self.mode = mode
        self.setup_function = setup_function
        self.module_filename = module_filename
        self.my_module_config = my_module_config
        self.execution_time = execution_time
        self.results_monthly = results_monthly
        self.results_hourly = results_hourly
        self.results_cumulative = results_cumulative
        self.results_daily = results_daily
        self.post_processing_options = simulation_parameters.post_processing_options
        self.kpi_collection_dict = kpi_collection_dict

        log.information("Selected " + str(len(self.post_processing_options)) + " post processing options:")
        for option in self.post_processing_options:
            log.information("Selected post processing option: " + str(option))



"""Dummy class and configuration returning electricity prices (from grid) and returns (injection) in each time step. """

# clean

import os
from typing import List, Any, Optional
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import pandas as pd
import numpy as np

# Owned
from hisim import component as cp
from hisim.simulationparameters import SimulationParameters
from hisim import utils
from hisim import loadtypes as lt
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class PriceSignalConfig(cp.ConfigBase):
    """Price Signal config class."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return PriceSignal.get_full_classname()

    building_name: str
    #: name of the price signal
    name: str
    country: str
    pricing_scheme: str
    installed_capacity: float
    price_signal_type: str
    fixed_price: list
    static_tou_price: list
    price_injection: float
    predictive_control: bool
    prediction_horizon: Optional[int]

    @classmethod
    def get_default_price_signal_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Default configuration for price signal."""
        config = PriceSignalConfig(
            building_name=building_name,
            name="PriceSignal",
            country="Germany",
            pricing_scheme="fixed",
            installed_capacity=10e3,
            price_signal_type="dummy",
            fixed_price=[],
            static_tou_price=[],
            price_injection=0.0,
            predictive_control=False,
            prediction_horizon=None,
        )
        return config


class PriceSignal(cp.Component):
    """Price Signal class.

    Class component that provides price for electricity.
    Outputs: Price for injection: cents/kWh, Price for purchase: cents/kWh
    """

    # Forecasts
    Price_Injection_Forecast_24h = "Price_Injection_Forecast_24h"
    Price_Purchase_Forecast_24h = "Price_Purchase_Forecast_24h"

    # Outputs
    PricePurchase = "PricePurchase"
    PriceInjection = "PriceInjection"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: PriceSignalConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialization of Price Signal class.

        :param my_simulation_parameters: _description_
        :type my_simulation_parameters: SimulationParameters
        :param config: _description_
        :type config: PriceSignalConfig
        """
        self.price_signal_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build_dummy(
            start=int(10 * 3600 / my_simulation_parameters.seconds_per_timestep),
            end=int(16 * 3600 / my_simulation_parameters.seconds_per_timestep),
        )

        # Outputs
        self.price_purchase_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PricePurchase,
            lt.LoadTypes.PRICE,
            lt.Units.EUR_PER_KWH,
            postprocessing_flag=[
                lt.LoadTypes.PRICE,
                lt.InandOutputType.ELECTRICITY_CONSUMPTION,
            ],
            output_description=f"here a description for {self.PricePurchase} will follow.",
        )
        self.price_injection_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PriceInjection,
            lt.LoadTypes.PRICE,
            lt.Units.EUR_PER_KWH,
            postprocessing_flag=[
                lt.LoadTypes.PRICE,
                lt.InandOutputType.ELECTRICITY_INJECTION,
            ],
            output_description=f"here a description for {self.PriceInjection} will follow.",
        )

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Outputs price signal of time step."""
        priceinjectionforecast = [0.1]
        pricepurchaseforecast = [0.5]
        if self.config.predictive_control and self.config.prediction_horizon:
            priceinjectionforecast = [0.1] * int(
                self.config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
            pricepurchaseforecast = [0.5] * int(
                self.config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
        elif self.price_signal_config.price_signal_type == "Prices at second half of 2021":
            priceinjectionforecast = [self.price_signal_config.price_injection] * int(
                self.config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
        elif self.price_signal_config.pricing_scheme == "dynamic":
            pricepurchaseforecast = self.price_signal_config.static_tou_price
        elif self.price_signal_config.pricing_scheme == "fixed":
            pricepurchaseforecast = self.price_signal_config.fixed_price
        elif self.price_signal_config.price_signal_type == "dummy":
            priceinjectionforecast = [10] * int(
                self.config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
            pricepurchaseforecast = [50] * int(
                self.config.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
            # pricepurchaseforecast = [ ]
            # for step in range( self.day ):
            #     x = timestep % self.day
            #     if x > self.start and x < self.end:
            #         pricepurchaseforecast.append( 20 * self.my_simulation_parameters.seconds_per_timestep / 3.6e6 )
            #     else:
            #         pricepurchaseforecast.append( 50 * self.my_simulation_parameters.seconds_per_timestep / 3.6e6 )
        else:
            priceinjectionforecast = [0.1]
            pricepurchaseforecast = [0.5]

        SingletonSimRepository().set_entry(
            key=SingletonDictKeyEnum.PRICEINJECTIONFORECAST24H,
            entry=priceinjectionforecast,
        )
        SingletonSimRepository().set_entry(
            key=SingletonDictKeyEnum.PRICEPURCHASEFORECAST24H,
            entry=pricepurchaseforecast,
        )
        stsv.set_output_value(self.price_purchase_channel, pricepurchaseforecast[0])
        stsv.set_output_value(self.price_injection_channel, priceinjectionforecast[0])

    def build_dummy(self, start: int, end: int) -> None:
        """Initialization of information if step function is used for prices."""
        self.start = start
        self.end = end

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        price_purchase = pd.read_csv(
            os.path.join(utils.HISIMPATH["price_signal"]["PricePurchase"]),
            index_col=0,
        )
        feed_in_tarrif = pd.read_csv(
            os.path.join(utils.HISIMPATH["price_signal"]["FeedInTarrif"]),
            index_col=0,
        )

        if "Fixed_Price_" + self.price_signal_config.country in price_purchase:
            self.price_signal_config.price_signal_type = "Prices at second half of 2021"
            fixed_price = price_purchase["Fixed_Price_" + self.price_signal_config.country].tolist()
            # convert euro/kWh to cent/kW-timestep
            p_conversion = 100 / (1000 * 3600 / self.my_simulation_parameters.seconds_per_timestep)
            fixed_price = [element * p_conversion for element in fixed_price]
            self.price_signal_config.fixed_price = np.repeat(
                fixed_price,
                int(3600 / self.my_simulation_parameters.seconds_per_timestep),
            ).tolist()

            static_tou_price = price_purchase["Static_TOU_Price_" + self.price_signal_config.country].tolist()
            static_tou_price = [element * p_conversion for element in static_tou_price]
            self.price_signal_config.static_tou_price = np.repeat(
                static_tou_price,
                int(3600 / self.my_simulation_parameters.seconds_per_timestep),
            ).tolist()

            fit_data = feed_in_tarrif.loc[self.price_signal_config.country]
            for i in range(len(fit_data)):
                if (
                    fit_data["min_capacity (kW)"].values[i] < self.price_signal_config.installed_capacity
                    and fit_data["max_capacity (kW)"].values[i] >= self.price_signal_config.installed_capacity
                ):
                    price_injection = fit_data["FIT"].values[i]
            self.price_signal_config.price_injection = price_injection * p_conversion
        pass

    def write_to_report(self) -> List[str]:
        """Writes relevant information to report."""
        return self.price_signal_config.get_string_dict()



"""Postprocessing: computes investment cost and CO2 footprint of technical equipment.

Functions from this file are called in Postprocessing option compute_kpis.
"""
# clean
from typing import List, Tuple
import pandas as pd
from hisim.components import (
    generic_hot_water_storage_modular,
    generic_pv_system,
    generic_smart_device,
    generic_heat_source,
    advanced_battery_bslib,
    generic_car,
    generic_chp,
    generic_hydrogen_storage,
    generic_electrolyzer,
)
from hisim.utils import HISIMPATH
from hisim.loadtypes import LoadTypes
from hisim.component_wrapper import ComponentWrapper


def read_in_component_costs() -> pd.DataFrame:
    """Reads data for cost and co2 emissions of component installation/investment from csv.

    :return: DataFrame with price and co2 footprint information of all relevant components.
    :rtype: pd.DataFrame
    """
    price_frame = pd.read_csv(HISIMPATH["component_costs"], sep=";", usecols=[0, 8, 9])
    price_frame.index = price_frame["Product/service"]  # type: ignore
    price_frame.drop(columns=["Product/service"], inplace=True)
    return price_frame


def compute_investment_cost(components: List[ComponentWrapper],) -> Tuple[float, float]:
    """Iterates over all components and computes annual investment cost and annual C02 footprint respectively.

    :param components: List of all configured components in the HiSIM system setup.
    :type components: List[ComponentWrapper]
    :return: annual investment cost for considered equipment and annual C02 footprint.
    :rtype: Tuple[float, float]
    """
    # initialize values
    investment_cost = 0.0
    co2_emissions = 0.0
    price_frame = read_in_component_costs()

    for component in components:
        if isinstance(component.my_component, generic_smart_device.SmartDevice):
            column = price_frame.iloc[price_frame.index == "Washing machine (or domestic appliances in general)"]
            component_capacity = 1.0
        elif isinstance(component.my_component, generic_pv_system.PVSystem):
            column = price_frame.iloc[price_frame.index == "Photovoltaic panel"]
            component_capacity = component.my_component.pvconfig.power_in_watt * 1e-3
        elif isinstance(component.my_component, generic_heat_source.HeatSource):
            if component.my_component.config.fuel == LoadTypes.DISTRICTHEATING:
                column = price_frame.iloc[price_frame.index == "Biomass district heating system"]
            elif component.my_component.config.fuel == LoadTypes.GAS:
                column = price_frame.iloc[price_frame.index == "Gas boiler"]
            elif component.my_component.config.fuel == LoadTypes.OIL:
                column = price_frame.iloc[price_frame.index == "Oil boiler"]
            elif component.my_component.config.fuel == LoadTypes.ELECTRICITY:
                column = price_frame.iloc[price_frame.index == "Electric heating"]
            component_capacity = component.my_component.config.power_th * 1e-3
        elif isinstance(component.my_component, generic_hot_water_storage_modular.HotWaterStorage):
            column = price_frame.iloc[price_frame.index == "Hot Water tank"]
            component_capacity = component.my_component.volume
        elif isinstance(component.my_component, advanced_battery_bslib.Battery):
            column = price_frame.iloc[price_frame.index == "Lithium iron phosphate battery"]
            component_capacity = component.my_component.custom_battery_capacity_generic_in_kilowatt_hour
        elif isinstance(component.my_component, generic_car.Car):
            if component.my_component.config.fuel == LoadTypes.ELECTRICITY:
                column = price_frame.iloc[price_frame.index == "Electric vehicle"]
            elif component.my_component.config.fuel == LoadTypes.DIESEL:
                column = price_frame.iloc[price_frame.index == "Diesel vehicle"]
            component_capacity = 1.0
        elif isinstance(component.my_component, generic_chp.SimpleCHP):
            if component.my_component.config.use == LoadTypes.GAS:
                column = price_frame.iloc[price_frame.index == "Gas powered Combined Heat and Power"]
            elif component.my_component.config.use == LoadTypes.HYDROGEN:
                column = price_frame.iloc[price_frame.index == "Hydrogen fuelcell"]
            component_capacity = component.my_component.config.p_fuel * 1e-3
        elif isinstance(component.my_component, generic_hydrogen_storage.GenericHydrogenStorage):
            column = price_frame.iloc[price_frame.index == "Hydrogen Storage"]
            component_capacity = component.my_component.config.max_capacity
        elif isinstance(component.my_component, generic_electrolyzer.GenericElectrolyzer):
            column = price_frame.iloc[price_frame.index == "Electrolyzer"]
            component_capacity = component.my_component.config.max_power * 1e-3
        else:
            continue
        co2_emissions = co2_emissions + float(column["annual Footprint"].iloc[0]) * component_capacity
        investment_cost = investment_cost + float(column["annual cost"].iloc[0]) * component_capacity

    return investment_cost, co2_emissions



""" Contains all the chart classes. """
# clean
import gc
from typing import Any
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from hisim import log
from hisim.postprocessing.chartbase import Chart, ChartFontsAndSize
from hisim import utils
from hisim.postprocessing.report_image_entries import ReportImageEntry
from hisim.simulationparameters import FigureFormat

mpl.rcParams["agg.path.chunksize"] = 10000


class Carpet(Chart, ChartFontsAndSize):  # noqa: too-few-public-methods

    """Class for carpet plots."""

    def __init__(
        self,
        output: Any,
        component_name: str,
        units: Any,
        directory_path: str,
        time_correction_factor: float,
        output_description: str,
        figure_format: FigureFormat,
    ) -> None:
        """Initalizes a carpot plot."""
        super().__init__(
            output=output,
            component_name=component_name,
            chart_type="Carpet",
            units=units,
            directory_path=directory_path,
            time_correction_factor=time_correction_factor,
            output_description=output_description,
            figure_format=figure_format,
        )

    def plot(self, xdims: int, data: Any) -> ReportImageEntry:
        """Makes a carpet plot."""
        log.trace("starting carpet plots")
        ydims = int(len(data) / xdims)  # number of calculated timesteps per day
        y_steps_per_hour = int(ydims / 24)

        try:
            database = data.values.reshape(xdims, ydims)
        except ValueError:
            log.error("Carpet plot can only deal with data containing entire days")

        if np.max(np.abs(data.values)) > 1.5e3:
            database = database * 1e-3
            self.units = f"k{self.units}"

        plot_data = np.flip(database.transpose(), axis=0)

        fig = plt.figure(figsize=self.figsize, dpi=self.dpi)

        axis = fig.add_subplot(111)
        mycolors = "viridis"
        color_map = mpl.colormaps.get_cmap(mycolors)

        plot = axis.pcolormesh(plot_data, cmap=color_map)
        plt.colorbar(plot).set_label(self.units, fontsize=self.fontsize_label)

        y_ticks = np.arange(0, 25 * y_steps_per_hour, 6 * y_steps_per_hour).tolist()
        axis.set_yticks(y_ticks)
        y_ticks_labels = np.flip(list(range(0, 25, 6)), axis=0)
        axis.set_yticklabels([str(i) for i in y_ticks_labels])

        if xdims == 365:
            x_ticks = np.arange(15, 346, 30).tolist()
            axis.set_xticks(x_ticks)
            axis.set_xticklabels([str(i) for i in self.months_abbrev_uppercase])

        # optimizing fonts
        fig.autofmt_xdate(rotation=45)
        # setting axis of the plot
        axis.set_ylabel("Time of day [h]", fontsize=self.fontsize_label)
        axis.set_xlabel("Month of the year", fontsize=self.fontsize_label)
        plt.title(self.title, fontsize=self.fontsize_title)
        plt.xticks(fontsize=self.fontsize_ticks)
        plt.yticks(fontsize=self.fontsize_ticks)
        plt.tight_layout()
        log.trace("finished carpet plot: " + self.filepath)
        plt.savefig(self.filepath2)
        plt.close()
        return ReportImageEntry(
            category=None,
            output_description=self.output_description,
            component_output_folder_path=self.component_output_folder_path,
            file_path=self.filepath2,
            unit=self.units,
            component_name=self.component_name,
            output_type=self.output_type,
        )


class Line(Chart, ChartFontsAndSize):  # noqa: too-few-public-methods

    """Makes a line chart."""

    # @utils.measure_memory_leak
    def __init__(
        self,
        output: Any,
        component_name: str,
        units: Any,
        directory_path: str,
        time_correction_factor: float,
        output_description: str,
        figure_format: FigureFormat,
    ):
        """Initializes a line chart."""
        if output_description is None:
            raise ValueError("Output description was None for component " + component_name)

        super().__init__(
            output=output,
            component_name=component_name,
            chart_type="Line",
            units=units,
            directory_path=directory_path,
            time_correction_factor=time_correction_factor,
            output_description=output_description,
            figure_format=figure_format,
        )

    @utils.measure_memory_leak
    def plot(self, data: Any) -> ReportImageEntry:
        """Makes a line plot."""

        mpl.use("Agg")

        _fig, axis = plt.subplots(figsize=self.figsize, dpi=self.dpi)
        x_zero = data.index
        plt.xticks(fontsize=self.fontsize_ticks, rotation=20)
        plt.yticks(fontsize=self.fontsize_ticks)

        # Rescale values in case they are too high
        data, self.units = self.rescale_y_axis(y_values=data, units=self.units)

        plt.plot(x_zero, data, color="green", linewidth=1.0)
        plt.ylabel(f"[{self.units}]", fontsize=self.fontsize_label)
        plt.xlabel("Time", fontsize=self.fontsize_label)
        plt.grid()
        plt.title(self.title, fontsize=self.fontsize_title)
        axis.set_xlim(xmin=x_zero[0])
        plt.tight_layout()
        # plt.savefig(self.filepath)
        plt.savefig(self.filepath2)
        plt.cla()
        plt.clf()
        plt.close("all")
        del x_zero
        gc.collect(2)
        return ReportImageEntry(
            category=None,
            output_description=self.output_description,
            component_output_folder_path=self.component_output_folder_path,
            file_path=self.filepath2,
            unit=self.units,
            component_name=self.component_name,
            output_type=self.output_type,
        )


class BarChart(Chart, ChartFontsAndSize):  # noqa: too-few-public-methods

    """Makes Bar charts."""

    original_pv_sol = [
        385.66,
        484.01,
        981.05,
        1096.7,
        1157,
        1299.9,
        1415.3,
        1266.1,
        1075.8,
        714.44,
        422.51,
        366.83,
    ]

    def __init__(
        self,
        output: Any,
        component_name: str,
        units: Any,
        directory_path: str,
        time_correction_factor: float,
        output_description: str,
        figure_format: FigureFormat,
    ):
        """Initializes the classes."""
        super().__init__(
            output=output,
            component_name=component_name,
            chart_type="Bar",
            units=units,
            directory_path=directory_path,
            time_correction_factor=time_correction_factor,
            output_description=output_description,
            figure_format=figure_format,
        )
        self.filename = f"monthly_{self.output}{self.figure_format}"

    def plot(self, data: Any) -> ReportImageEntry:
        """Plots the bar chart."""
        width = 0.35
        # Specify the values of blue bars (height)

        # Position of bars on x-axis
        ind = np.arange(12)

        # Width of a bar
        width = 0.4

        # Rescale values in case they are too high
        data, self.units = self.rescale_y_axis(y_values=data, units=self.units)

        plt.subplots(figsize=self.figsize, dpi=self.dpi)
        plt.bar(ind, data, width)
        plt.xticks(
            ticks=ind, labels=[str(i) for i in self.months_abbrev_uppercase], fontsize=self.fontsize_ticks,
        )

        plt.yticks(fontsize=self.fontsize_ticks)
        plt.title(f"{self.title} Monthly", fontsize=self.fontsize_title)
        plt.grid()
        plt.tight_layout()
        plt.ylabel(f"[{self.units}]", fontsize=self.fontsize_label)

        plt.savefig(self.filepath2)
        plt.close()
        return ReportImageEntry(
            category=None,
            output_description=self.output_description,
            component_output_folder_path=self.component_output_folder_path,
            file_path=self.filepath2,
            unit=self.units,
            component_name=self.component_name,
            output_type=self.output_type,
        )



"""Advanced fuel cell controller module."""

# clean
from typing import Any
from math import ceil
from copy import deepcopy
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput, DisplayConfig
from hisim import loadtypes as lt

from hisim.components.configuration import (
    CHPControllerConfig,
    GasControllerConfig,
    AdvElectrolyzerConfig,
)
from hisim.components import advanced_fuel_cell as chp
from hisim.components.configuration import ExtendedControllerConfig
from hisim.simulationparameters import SimulationParameters
from hisim import log


class ExtendedControllerSimulation:
    """Extended Controller Simulation."""

    def __init__(self, config: ExtendedControllerConfig):
        """Initialize the class."""
        pass

    def regulate_chp_mode_power(
        self,
        state_chp: Any,
        runtime_chp: Any,
        power_supply_pv: Any,
        electricity_demand_household: Any,
        seconds_per_timestep: Any,
    ) -> Any:
        """Regulate the power.

        :param power_supply_pv:
        :param electricity_demand_household:
        :param seconds_per_timestep:
        :return:
        """
        demand = electricity_demand_household - power_supply_pv
        if chp.CHPConfig.is_modulating:
            power_states_possible = ExtendedControllerConfig.chp_power_states_possible
            min_power = chp.CHPConfig.p_el_min
            max_power = chp.CHPConfig.p_el_max
            power_delta = max_power - min_power
            # the first step is already the minimum power
            power_per_step_size = power_delta / (power_states_possible - 1)

            if demand <= 0:
                if state_chp == 0:
                    pass
                else:
                    minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                    if runtime_chp >= minimum_timesteps:
                        # switch off chp
                        state_chp = 0
                        runtime_chp = 0
                    else:
                        state_chp = 1 / power_states_possible
            elif 0 < demand < chp.CHPConfig.p_el_min:
                # maximum_autarky --> production time if P > 0 or production only if P in range?
                if ExtendedControllerConfig.maximum_autarky:
                    state_chp = 1 / power_states_possible  # eg 0.1
                else:
                    if state_chp == 0:
                        pass  # state_chp = 0
                    else:
                        minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                        if runtime_chp >= minimum_timesteps:
                            # switch off chp
                            state_chp = 0
                            runtime_chp = 0
                        else:
                            state_chp = 1 / power_states_possible
            elif chp.CHPConfig.p_el_min <= demand < chp.CHPConfig.p_el_max:
                # minimum power plus the demand depending power
                state_chp = (1 + ceil((demand - chp.CHPConfig.p_el_min) / power_per_step_size)) / power_states_possible
                assert (1 / power_states_possible) < state_chp <= 1
            else:  # demand >= chp.CHPConfig.P_el_max
                state_chp = 1

            if state_chp == 0:
                generated_electricity = 0
                power_from_or_to_grid = demand - generated_electricity
            elif 0 < state_chp <= 1:
                generated_electricity = chp.CHPConfig.p_el_min + power_per_step_size * (
                    state_chp * power_states_possible - 1
                )
                power_from_or_to_grid = demand - generated_electricity
            else:
                log.error("Wrong controller state")
                raise ValueError

        else:
            if demand <= 0:
                if state_chp == 0:
                    power_from_or_to_grid = demand
                else:
                    minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                    if runtime_chp >= minimum_timesteps:
                        state_chp = 0
                        power_from_or_to_grid = demand
                    else:
                        state_chp = 1
                        power_from_or_to_grid = demand - chp.CHPConfig.p_el_max

            else:  # demand > 0
                state_chp = 1
                power_from_or_to_grid = demand - chp.CHPConfig.p_el_max

        # independent of modulation
        if state_chp > 0:
            runtime_chp += 1
        else:
            runtime_chp = 0

        return state_chp, runtime_chp, power_from_or_to_grid

    def regulate_chp_mode_heat(
        self,
        temperatures_in_tank,
        previous_state,
        runtime_chp,
        pv_production,
        electricity_demand_household,
        seconds_per_timestep,
    ):
        """Regulate chp mode heat."""

        # the heat model has no modulation because there is a buffer (warm water storage)

        heights_in_tank = CHPControllerConfig.heights_in_tank

        if (CHPControllerConfig.height_upper_sensor or CHPControllerConfig.height_lower_sensor) not in heights_in_tank:
            log.error(
                "Wrong sensor setting. Only 0, 20, 40, 60, 80, 100% are allowed.\n"
                "You tried "
                + str(CHPControllerConfig.height_upper_sensor)
                + " and "
                + str(CHPControllerConfig.height_lower_sensor)
            )
            raise ValueError

        # get temperatures at the chosen sensors
        for i, height_in_tank in enumerate(heights_in_tank):
            if CHPControllerConfig.height_upper_sensor == height_in_tank:
                temperature_upper_sensor = temperatures_in_tank[i]
            if CHPControllerConfig.height_lower_sensor == height_in_tank:
                temperature_lower_sensor = temperatures_in_tank[i]
        state_chp = previous_state
        # upper sensor
        if temperature_upper_sensor < CHPControllerConfig.temperature_switch_on:
            # switch on
            state_chp = 1

        # lower sensor (no check needed if chp is off)
        if state_chp == 1:
            if temperature_lower_sensor > CHPControllerConfig.temperature_switch_off:
                minimum_timesteps = (CHPControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                if runtime_chp > minimum_timesteps:
                    # switch off is possible if chp has run at least xx min
                    state_chp = 0

        # increase counter
        if state_chp == 1:
            runtime_chp += 1
        elif state_chp == 0:
            runtime_chp = 0
        else:
            log.error("Wrong state_chp")
            raise ValueError
        # 'easy equation' --> no modulation
        power_from_or_to_grid = electricity_demand_household - pv_production - chp.CHPConfig.p_el_max * state_chp

        return state_chp, runtime_chp, power_from_or_to_grid

    def regulate_gas_heater(
        self,
        temperatures_in_tank: Any,
        previous_state: float,
        runtime_counter: int,
        seconds_per_timestep: int,
    ) -> Any:
        """Regulate gas heater."""

        # the gas_heater model has no modulation because there is a buffer (warm water storage)
        # ToDo: future --> make modulating possible if the waste energy is to high? reduce power, increase mass_flow

        heights_in_tank = CHPControllerConfig.heights_in_tank
        if (GasControllerConfig.height_upper_sensor or GasControllerConfig.height_lower_sensor) not in heights_in_tank:
            log.error(
                "Wrong sensor setting. Only 0, 20, 40, 60, 80, 100% are allowed.\n"
                "You tried "
                + str(GasControllerConfig.height_upper_sensor)
                + " and "
                + str(GasControllerConfig.height_lower_sensor)
            )
            raise ValueError

        for i, height_in_tank in enumerate(heights_in_tank):
            if GasControllerConfig.height_upper_sensor == height_in_tank:
                temperature_upper_sensor = temperatures_in_tank[i]
            if GasControllerConfig.height_lower_sensor == height_in_tank:
                temperature_lower_sensor = temperatures_in_tank[i]
        state_gas_heater = previous_state
        # upper sensor
        if temperature_upper_sensor < GasControllerConfig.temperature_switch_on:
            # switch on
            state_gas_heater = 1

        # lower sensor (no check needed if gas heater is off)
        if state_gas_heater == 1:
            if temperature_lower_sensor > GasControllerConfig.temperature_switch_off:
                minimum_timesteps = (GasControllerConfig.minimum_runtime_minutes * 60) / seconds_per_timestep
                if runtime_counter > minimum_timesteps:
                    # switch off is possible if gas_heater has run at least xx min
                    state_gas_heater = 0
            else:
                state_gas_heater = 1

        # increase counter
        if state_gas_heater == 1:
            runtime_counter += 1
        elif state_gas_heater == 0:
            runtime_counter = 0
        else:
            log.error("Wrong state_chp")
            raise ValueError

        return state_gas_heater, runtime_counter

    def power_distribution_to_electrolyzer(self, power_from_or_to_grid: Any) -> Any:
        """Power distribution to electrolyzer.

        Hydrogen storage must be able to store the produced massflow of hydrogen.
        Otherwise the dimensioning of the system is incorrect --> is checked in hydrogen_storage
        power_from_or_to_grid is positive if there is a demand of energy and negative if there is a surplus of energy.
        power_available is defined the other way round --> negate the value
        """
        power_available = -power_from_or_to_grid

        if power_available < AdvElectrolyzerConfig.min_power:
            power_to_electrolyzer = 0
            # no change
            # power_from_or_to_grid = power_available
        elif AdvElectrolyzerConfig.min_power <= power_available <= AdvElectrolyzerConfig.max_power:
            power_to_electrolyzer = power_available
            power_from_or_to_grid = 0
        else:  # power_available > ElectrolyzerConfig.max_power:
            power_to_electrolyzer = AdvElectrolyzerConfig.max_power
            # not al the electricity can go to the electolyzer --> power_from_or_to_gridstays negative
            power_from_or_to_grid = power_to_electrolyzer - power_available

        return power_to_electrolyzer, power_from_or_to_grid


class ExtendedController(Component):
    """Extended Controller class."""

    # inputs
    ElectricityDemand = "Electricity Demand"  # W
    PV_Production = "PV Production"  # W

    # temperatures (input)
    Temperature0Percent = "Temperature 0 Percent"  # °C
    Temperature20Percent = "Temperature 20 Percent"  # °C
    Temperature40Percent = "Temperature 40 Percent"  # °C
    Temperature60Percent = "Temperature 60 Percent"  # °C
    Temperature80Percent = "Temperature 80 Percent"  # °C
    Temperature100Percent = "Temperature 100 Percent"  # °C

    # Output
    ControllerCHP = "Controller CHP"
    ControllerGasHeater = "Controller Gas Heater"
    PowerToElectrolyzer = "Power To Electrolyzer"
    PowerFromOrToGrid = "Power From Or To Grid"

    RuntimeCounterCHP = "RuntimeCounterCHP"
    RuntimeCounterGasHeater = "RuntimeCounterGasHeater"

    def __init__(
        self,
        component_name: str,
        config: ExtendedControllerConfig,
        my_simulation_parameters: SimulationParameters,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        # Input
        self.electricity_demand_household_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.ElectricityDemand,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )
        self.pv_production_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.PV_Production,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )

        self.temperature_0_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature0Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )
        self.temperature_20_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature20Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )
        self.temperature_40_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature40Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )
        self.temperature_60_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature60Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )
        self.temperature_80_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature80Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )
        self.temperature_100_percent_channel: ComponentInput = self.add_input(
            self.component_name,
            ExtendedController.Temperature100Percent,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            True,
        )

        # Output
        self.controller_chp_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.ControllerCHP,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
        )
        self.controller_gas_heater_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.ControllerGasHeater,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
        )
        self.power_to_electrolyzer_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.PowerToElectrolyzer,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
        )
        self.power_from_or_to_grid_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.PowerFromOrToGrid,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
        )

        self.runtime_counter_chp_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.RuntimeCounterCHP,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
        )
        self.runtime_counter_gas_heater_channel: ComponentOutput = self.add_output(
            self.component_name,
            ExtendedController.RuntimeCounterGasHeater,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
        )

        self.extended_controller = ExtendedControllerSimulation(config)
        self.seconds_per_timestep = my_simulation_parameters.seconds_per_timestep

        # CHP state/runtime & Gas state/runtime
        self.state_chp1 = 0
        self.runtime_chp1 = 0
        self.state_gas_heater1 = 0
        self.runtime_gas_heater1 = 0

        # self.previous_state = [self.state_chp1, self.runtime_chp1, self.state_gas_heater1, self.runtime_gas_heater1]
        self.previous_state_chp1 = self.state_chp1
        self.previous_runtime_chp1 = self.runtime_chp1
        self.previous_state_gas_heater1 = self.state_gas_heater1
        self.previous_runtime_gas_heater1 = self.runtime_gas_heater1

        self.test_pv: float = 0
        self.test_grid: float = 0
        self.test_electrolyzer: float = 0
        self.test_demand: float = 0
        self.test_state: float = 0

    def i_save_state(self) -> None:
        """Saves the state."""
        # self.previous_state = self.extended_controller.begin_new_timestep()
        self.previous_state_chp1 = deepcopy(self.state_chp1)
        self.previous_runtime_chp1 = deepcopy(self.runtime_chp1)
        self.previous_state_gas_heater1 = deepcopy(self.state_gas_heater1)
        self.previous_runtime_gas_heater1 = deepcopy(self.runtime_gas_heater1)

    def i_restore_state(self) -> None:
        """Restores the state."""
        # self.extended_controller.reset_to_last_timestep(self.previous_state)
        self.state_chp1 = deepcopy(self.previous_state_chp1)
        self.runtime_chp1 = deepcopy(self.previous_runtime_chp1)
        self.state_gas_heater1 = deepcopy(self.previous_state_gas_heater1)
        self.runtime_gas_heater1 = deepcopy(self.previous_runtime_gas_heater1)

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the state."""
        if force_convergence:
            return

        # Inputs
        electricity_demand_household = stsv.get_input_value(self.electricity_demand_household_channel)
        pv_production: float = stsv.get_input_value(self.pv_production_channel)

        # not needed for power
        if ExtendedControllerConfig.chp_mode == "heat" or ExtendedControllerConfig.gas_heater:
            temperature_0_percent = stsv.get_input_value(self.temperature_0_percent_channel)
            temperature_20_percent = stsv.get_input_value(self.temperature_20_percent_channel)
            temperature_40_percent = stsv.get_input_value(self.temperature_40_percent_channel)
            temperature_60_percent = stsv.get_input_value(self.temperature_60_percent_channel)
            temperature_80_percent = stsv.get_input_value(self.temperature_80_percent_channel)
            temperature_100_percent = stsv.get_input_value(self.temperature_100_percent_channel)

            temperatures_in_tank = [
                temperature_0_percent,
                temperature_20_percent,
                temperature_40_percent,
                temperature_60_percent,
                temperature_80_percent,
                temperature_100_percent,
            ]

        # Combined heat and power plant
        if ExtendedControllerConfig.chp:
            if ExtendedControllerConfig.chp_mode == "power":
                # chp is running depending on the delta between pv_production and electricity_demand_household
                (
                    self.state_chp1,
                    self.runtime_chp1,
                    power_from_or_to_grid,
                ) = self.extended_controller.regulate_chp_mode_power(
                    self.state_chp1,
                    self.runtime_chp1,
                    pv_production,
                    electricity_demand_household,
                    self.seconds_per_timestep,
                )
            elif ExtendedControllerConfig.chp_mode == "heat":
                (
                    self.state_chp1,
                    self.runtime_chp1,
                    power_from_or_to_grid,
                ) = self.extended_controller.regulate_chp_mode_heat(
                    temperatures_in_tank,
                    self.state_chp1,
                    self.runtime_chp1,
                    pv_production,
                    electricity_demand_household,
                    self.seconds_per_timestep,
                )
            else:
                log.error("Wrong chp controller settings! Choose between heat and power")
                raise ValueError
        else:
            power_from_or_to_grid = pv_production - electricity_demand_household

        # Gas heater
        if ExtendedControllerConfig.gas_heater:
            (
                self.state_gas_heater1,
                self.runtime_gas_heater1,
            ) = self.extended_controller.regulate_gas_heater(
                temperatures_in_tank,
                self.state_gas_heater1,
                self.runtime_gas_heater1,
                self.seconds_per_timestep,
            )

        if not ExtendedControllerConfig.chp or not ExtendedControllerConfig.gas_heater:
            log.error("Choose a energy source")
            raise ValueError

        # Electrolyzer
        # ToDo: Sollte er laufen wenn die KWK auch läuft... (stromgeführt nein, wärmegefüht evtl ja slange keine Batterie da ist)

        if ExtendedControllerConfig.electrolyzer:
            (
                power_to_electrolyzer,
                power_from_or_to_grid,
            ) = self.extended_controller.power_distribution_to_electrolyzer(power_from_or_to_grid)
        else:
            power_to_electrolyzer = 0

        # Outputs
        stsv.set_output_value(self.controller_chp_channel, self.state_chp1)
        # stsv.set_output_value(self.controller_gas_heater, self.extended_controller.state_gas_heater)
        stsv.set_output_value(self.controller_gas_heater_channel, self.state_gas_heater1)

        stsv.set_output_value(self.power_to_electrolyzer_channel, power_to_electrolyzer)
        stsv.set_output_value(self.power_from_or_to_grid_channel, power_from_or_to_grid)

        stsv.set_output_value(self.runtime_counter_chp_channel, self.runtime_chp1)
        stsv.set_output_value(self.runtime_counter_gas_heater_channel, self.runtime_gas_heater1)

        # test
        self.test_demand = electricity_demand_household
        self.test_pv = pv_production
        self.test_grid = power_from_or_to_grid
        self.test_electrolyzer = power_to_electrolyzer
        self.test_state = self.state_chp1

        # self.test_state = self.extended_controller.state_chp

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        # check the electricity balance
        if chp.CHPConfig.is_modulating:
            if self.test_state > 0:
                power_chp_test = chp.CHPConfig.p_el_min + (chp.CHPConfig.p_el_max - chp.CHPConfig.p_el_min) / (
                    ExtendedControllerConfig.chp_power_states_possible - 1
                ) * (self.test_state * ExtendedControllerConfig.chp_power_states_possible - 1)
            else:
                power_chp_test = 0

            if 0.00001 > (self.test_pv + power_chp_test + self.test_grid - self.test_demand - self.test_electrolyzer):
                pass
            else:
                log.error("Wrong energy balance:")
                log.error("State CHP: " + str(self.test_state))
                log.error("test_pv: " + str(self.test_pv))
                log.error("power_chp_test: " + str(power_chp_test))
                log.error("test_grid: " + str(self.test_grid))
                log.error("test_demand: " + str(self.test_demand))
                log.error("test_electrolyzer: " + str(self.test_electrolyzer))
                raise ValueError

        else:
            if 0.00001 > (
                self.test_pv
                + chp.CHPConfig.p_el_max * self.test_state
                + self.test_grid
                - self.test_electrolyzer
                - self.test_demand
            ):
                pass
            else:
                log.error("Wrong energy balance:")
                log.error("State CHP: " + str(self.test_state))
                log.error("test_pv: " + str(self.test_pv))
                log.error("power_chp_test: " + str(chp.CHPConfig.p_el_max))
                log.error("test_grid: " + str(self.test_grid))
                log.error("test_demand: " + str(self.test_demand))
                log.error("test_electrolyzer: " + str(self.test_electrolyzer))
                raise ValueError



"""Build and simulate a system setup for a specific system setup that is defined in a JSON file.

Result files are stored in `/results`.
See `tests/test_system_setup_starter.py` for a system setup.

Run `hisim/system_setup_starter.py <json-file>` to start a simulation.
Required fields in the JSON file are: `path_to_module`, `function_in_module` and
`simulation_parameters`. SimulationParameters from the system_setups is not used. Instead, the
parameters from the JSON are set.

Optional field: `building_config`
The values from `building_config` replace single values of the system setup's building object. When
present, it is used to scale the default configuration with `get_scaled_default()`.

Optional field: `system_setup_config`
The values from `system_setup_config` overwrite specific values of the configuration object.
Arguments that are not present keep the (scaled) default value.
"""

import json
import datetime
from pathlib import Path
from typing import Union, Tuple, Optional
from copy import deepcopy

from hisim import log
from hisim.hisim_main import main
from hisim.utils import set_attributes_of_dataclass_from_dict
from hisim.simulator import SimulationParameters


def make_system_setup(
    parameters_json: Union[dict, list], result_directory: str
) -> Tuple[str, Optional[SimulationParameters], str]:
    """Read setup parameters from JSON and build a system setup for a specific system setup.

    The setup is simulated and result files are stored in `result_directory`.
    """
    if isinstance(parameters_json, list):
        raise NotImplementedError("System Setup Starter can only handle one setup at a time for now.")

    _parameters_json = deepcopy(parameters_json)
    Path(result_directory).mkdir(parents=True, exist_ok=True)  # pylint: disable=unexpected-keyword-arg
    path_to_module = _parameters_json.pop("path_to_module")

    simulation_parameters_dict = _parameters_json.pop("simulation_parameters")
    module_config_path = str(Path(result_directory).joinpath("module_config.json"))
    simulation_parameters_path = str(Path(result_directory).joinpath("simulation_parameters.json"))
    options = _parameters_json.pop("options", {})
    building_config = _parameters_json.pop("building_config", {})
    system_setup_config = _parameters_json.pop("system_setup_config", {})
    module_config_dict = {
        "options": options,
        "building_config": building_config,
        "system_setup_config": system_setup_config,
    }

    if _parameters_json:
        raise AttributeError(f"There are unused attributes ({_parameters_json.keys()}) in parameters JSON.")

    # Set custom simulation parameters
    simulation_parameters = SimulationParameters(
        start_date=datetime.datetime.fromisoformat(simulation_parameters_dict.pop("start_date")),
        end_date=datetime.datetime.fromisoformat(simulation_parameters_dict.pop("end_date")),
        seconds_per_timestep=simulation_parameters_dict.pop("seconds_per_timestep"),
        result_directory=result_directory,
    )
    set_attributes_of_dataclass_from_dict(simulation_parameters, simulation_parameters_dict)

    with open(simulation_parameters_path, "w", encoding="utf8") as out_file:
        out_file.write(simulation_parameters.to_json())  # ignore: type

    with open(module_config_path, "w", encoding="utf8") as out_file:
        out_file.write(json.dumps(module_config_dict))  # ignore: type

    return (
        path_to_module,
        simulation_parameters,
        module_config_path,
    )


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        PARAMETERS_JSON_FILE = "input/request.json"
        RESULT_DIRECTORY = "results"
        if not Path(PARAMETERS_JSON_FILE).is_file():
            log.information("Please specify an input JSON file or place it in `input/request.json`.")
            sys.exit(1)
    elif len(sys.argv) == 2:
        PARAMETERS_JSON_FILE = sys.argv[1]
        RESULT_DIRECTORY = "results"
    elif len(sys.argv) == 3:
        PARAMETERS_JSON_FILE = sys.argv[1]
        RESULT_DIRECTORY = sys.argv[2]
    else:
        log.information("HiSim from JSON received too many arguments.")
        RESULT_DIRECTORY = ""
        sys.exit(1)

    log.information(f"Reading parameters from {PARAMETERS_JSON_FILE}.")
    with open(PARAMETERS_JSON_FILE, "r", encoding="utf8") as file:
        my_parameters_json: Union[dict, list] = json.load(file)

    (
        my_path_to_module,
        my_simulation_parameters,
        my_module_config,
    ) = make_system_setup(parameters_json=my_parameters_json, result_directory=RESULT_DIRECTORY)

    main(
        my_path_to_module,
        my_simulation_parameters,
        my_module_config,
    )
